from io import StringIO
import textwrap
import os
import tempfile
import numpy

from amuse import io
from amuse.test import amusetest
from amuse.io import text
from amuse.units import units
from amuse.units import quantities, core
from amuse.units import generic_unit_system
from amuse import datamodel


class CursorTests(amusetest.TestCase):

    def test1(self):
        contents = "1\n2\n3"
        data_file = StringIO(contents)
        instance = text.LineBasedFileCursor(data_file)

        self.assertEqual("1", instance.line())
        self.assertEqual("1", instance.line())
        instance.forward()
        self.assertEqual("2", instance.line())
        instance.forward()
        self.assertEqual("3", instance.line())
        self.assertEqual("3", instance.line())

    def test2(self):
        contents = "1\n2\n3"
        data_file = StringIO(contents)
        instance = text.LineBasedFileCursor(data_file)

        self.assertFalse(instance.is_at_end())
        instance.forward()
        self.assertFalse(instance.is_at_end())
        instance.forward()
        self.assertFalse(instance.is_at_end())
        instance.forward()
        self.assertTrue(instance.is_at_end())


class TableFormattedTextTests(amusetest.TestCase):

    def test1(self):
        contents = "#header\n1 2 3\n4 5 6\n       \n7 8 9\n     "
        data_file = StringIO(contents)
        instance = text.TableFormattedText("test.txt", data_file)
        instance.attribute_names = ['a', 'b', 'c']
        particles = instance.load()

        self.assertEqual(len(particles), 3)
        self.assertEqual(particles[0].a, 1)
        self.assertFalse(quantities.is_quantity(particles[0].a))

    def test2(self):
        p = datamodel.Particles(2)
        p.a = [1, 4] | units.m
        p.b = [2, 5] | units.m
        p.c = [3, 6] | units.m

        data_file = StringIO()
        instance = text.TableFormattedText("test.txt", data_file, p)
        instance.attribute_types = [units.m, units.m, units.m]
        instance.store()

        contents = data_file.getvalue()
        self.assertEqual("#a b c\n#m m m\n1.0 2.0 3.0\n4.0 5.0 6.0\n", contents)

    def test3(self):
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.MSun
        x.radius = [3.0, 4.0] | units.RSun
        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            io.write_set_to_file(x, filename, "txt", attribute_types=(units.MSun, units.RSun))
            with open(filename, "r") as f:
                contents = f.read()
        self.assertEqual("#mass radius\n#MSun RSun\n1.0 3.0\n2.0 4.0\n", contents)

    def test4(self):
        mass = [1.0, 2.0, 3.0] | generic_unit_system.mass
        length = [3.0, 4.0, 5.0] | generic_unit_system.length

        stream = StringIO()
        output = text.TableFormattedText(stream=stream)
        output.quantities = (mass, length)
        output.store()
        contents = stream.getvalue()
        self.assertEqual("#col(0) col(1)\n#mass length\n1.0 3.0\n2.0 4.0\n3.0 5.0\n", contents)

    def test5(self):
        x = datamodel.Particles(2)
        x.mass = [1.0, 2.0] | units.MSun
        x.radius = [3.0, 4.0] | units.RSun

        expected = [
            "#mass radius\n#MSun RSun\n{0} 1.0 3.0\n{1} 2.0 4.0\n".format(x[0].key, x[1].key),
            "#mass radius\n#MSun RSun\n1.0 {0} 3.0\n2.0 {1} 4.0\n".format(x[0].key, x[1].key),
            "#mass radius\n#MSun RSun\n1.0 3.0 {0}\n2.0 4.0 {1}\n".format(x[0].key, x[1].key),
        ]
        for column_index, expected_content in enumerate(expected):
            with tempfile.NamedTemporaryFile() as tmp:
                filename = tmp.name
                io.write_set_to_file(
                    x,
                    filename,
                    "txt",
                    key_in_column=column_index,
                    attribute_types=(units.MSun, units.RSun)
                )

                with open(filename, "r") as f:
                    contents = f.read()

                self.assertEqual(expected_content, contents)

                y = io.read_set_from_file(
                    filename,
                    "txt",
                    key_in_column=column_index,
                    attribute_types=(units.MSun, units.RSun),
                    attribute_names=('mass', 'radius')
                )

            self.assertEqual(y[0], x[0])
            self.assertEqual(y[1], x[1])

    def test6(self):
        p = datamodel.Particles(2)
        p.a = [1., 4.]
        p.b = [2, 5] | units.m
        p.c = [3, 6] | units.m

        data_file = StringIO()
        instance = text.TableFormattedText("test.txt", data_file, p)
        instance.attribute_types = [None, units.m, units.m]
        instance.store()

        contents = data_file.getvalue()
        self.assertEqual("#a b c\n#- m m\n1.0 2.0 3.0\n4.0 5.0 6.0\n", contents)

    def test7(self):
        p = datamodel.Particles(2)
        p.a = [1.0, 4.0]
        p.b = [2, 5] | units.m
        p.c = [3, 6] | units.m

        data_file = StringIO()
        instance = text.TableFormattedText("test.txt", data_file, p)
        instance.store()

        contents = data_file.getvalue()
        self.assertEqual("#a b c\n#- m m\n1.0 2.0 3.0\n4.0 5.0 6.0\n", contents)

    def test8(self):
        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            table = io.ReportTable(
                filename,
                "txt",
                attribute_types=(units.MSun, units.RSun)
            )
            table.add_row(1.0 | units.MSun, 3.0 | units.RSun)
            table.add_row(2.0 | units.MSun, 4.0 | units.RSun)
            table.close()

            with open(filename, "r") as f:
                contents = f.read()
        self.assertEqual("#MSun RSun\n1.0 3.0\n2.0 4.0\n", contents)

    def test9(self):
        p = datamodel.Particles(5)
        p.a = [1.0, 2.0, 3.0, 4.0, 5.0]
        p.b = [10, 11, 12, 13, 14] | units.m
        p.c = [20, 21, 22, 23, 24] | units.m

        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            io.write_set_to_file(
                p,
                filename,
                "txt",
                attribute_names=('a', 'b', 'c'),
                attribute_types=(None, units.m, units.m),
                maximum_number_of_lines_buffered=1,
            )
            with open(filename, "r") as f:
                contents = f.read()

            expected_contents = '#a b c\n#- m m\n1.0 10.0 20.0\n2.0 11.0 21.0\n3.0 12.0 22.0\n4.0 13.0 23.0\n5.0 14.0 24.0\n'
            self.assertEqual(expected_contents, contents)
            p2 = io.read_set_from_file(
                filename,
                "txt",
                attribute_names=('a', 'b', 'c'),
                attribute_types=(None, units.m, units.m),
                maximum_number_of_lines_buffered=1,
            )
        self.assertAlmostRelativeEquals(p2.a, p.a)
        self.assertAlmostRelativeEquals(p2.b, p.b)
        self.assertAlmostRelativeEquals(p2.c, p.c)

    def test10(self):
        p = datamodel.Particles(keys=[30, 31, 32, 33, 34])
        p.a = [1.0, 2.0, 3.0, 4.0, 5.0]
        p.b = [10, 11, 12, 13, 14] | units.m
        p.c = [20, 21, 22, 23, 24] | units.m
        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            io.write_set_to_file(
                p,
                filename,
                "txt",
                attribute_names=('a', 'b', 'c'),
                attribute_types=(None, units.m, units.m),
                maximum_number_of_lines_buffered=1,
                key_in_column=0
            )
            with open(filename, "r") as f:
                contents = f.read()
            expected_contents = '#a b c\n#- m m\n30 1.0 10.0 20.0\n31 2.0 11.0 21.0\n32 3.0 12.0 22.0\n33 4.0 13.0 23.0\n34 5.0 14.0 24.0\n'
            self.assertEqual(expected_contents, contents)
            p2 = io.read_set_from_file(
                filename,
                "txt",
                attribute_names=('a', 'b', 'c'),
                attribute_types=(None, units.m, units.m),
                maximum_number_of_lines_buffered=1,
                key_in_column=0
            )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)
        self.assertAlmostRelativeEquals(p2.b, p.b)
        self.assertAlmostRelativeEquals(p2.c, p.c)

    def test11(self):
        p = datamodel.Particles(200)
        p.a = 2 | units.m

        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            io.write_set_to_file(
                p,
                filename,
                "txt",
                attribute_names=('a'),
                attribute_types=(units.m,),
                maximum_number_of_lines_buffered=10,
                key_in_column=0
            )
            p2 = io.read_set_from_file(
                filename,
                "txt",
                attribute_names=('a'),
                attribute_types=(units.m,),
                maximum_number_of_lines_buffered=10,
                key_in_column=0
            )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)

    def test12(self):
        print("Test Text IO with specific data types (string, int, float)")
        daltons = datamodel.Particles(keys=[30, 31, 32, 33])
        daltons.name = ["Joe", "William", "Jack", "Averell"]
        daltons.length = [1.1, 1.4, 1.7, 2.0] | core.unit_with_specific_dtype(units.m, "float32")
        daltons.age = [21, 20, 19, 18] | core.unit_with_specific_dtype(units.yr, "int32")
        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            path = os.path.abspath(os.path.join(self.get_path_to_results(), filename))
            io.write_set_to_file(
                daltons,
                path,
                "txt",
                attribute_names=('name', 'length', 'age'),
                attribute_types=(None, units.m, units.yr),
                maximum_number_of_lines_buffered=2,
                key_in_column=0
            )
            with open(path, "r") as f:
                contents = f.read()
            expected_contents = '#name length age\n#- m yr\n30 Joe 1.1 21\n31 William 1.4 20\n32 Jack 1.7 19\n33 Averell 2.0 18\n'
            self.assertEqual(expected_contents, contents)
    
            read = io.read_set_from_file(
                path,
                "txt",
                attribute_names=('name', 'length', 'age'),
                attribute_types=(None, units.m, units.yr),
                attribute_dtypes=("str", "float32", "int32"),
                maximum_number_of_lines_buffered=2,
                key_in_column=0
            )
        self.assertEqual(read.key, daltons.key)
        self.assertEqual(read.name, daltons.name)
        self.assertEqual(read.length, daltons.length)
        self.assertEqual(read.age, daltons.age)
        self.assertTrue(read.name.dtype.kind in ["S", "U"])
        self.assertEqual(str(read.length.value_in(units.m).dtype), "float32")
        self.assertEqual(str(read.age.value_in(units.yr).dtype), "int32")

    def test13(self):
        p = datamodel.Particles(100)
        p.a = numpy.arange(0, 1, 0.01) | units.m

        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            path = os.path.abspath(os.path.join(self.get_path_to_results(), filename))

            io.write_set_to_file(
                p,
                path,
                "amuse-txt",
                attribute_names=('a'),
                maximum_number_of_lines_buffered=10,
                key_in_column=0
            )
            p2 = io.read_set_from_file(
                path,
                "txt",
                maximum_number_of_lines_buffered=10,
                key_in_column=0
            )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)

    def test14(self):
        p = datamodel.Particles(100)
        p.a = numpy.arange(0, 1, 0.01) | units.m
        p.b = numpy.arange(0, 1, 0.01)

        with tempfile.NamedTemporaryFile() as tmp:
            filename = tmp.name
            io.write_set_to_file(
                p,
                filename,
                "amuse-txt",
                maximum_number_of_lines_buffered=10,
                key_in_column=0
            )
            p2 = io.read_set_from_file(
                filename,
                "txt",
                maximum_number_of_lines_buffered=10,
                key_in_column=0
            )
        self.assertEqual(p2.key, p.key)
        self.assertAlmostRelativeEquals(p2.a, p.a)
        self.assertAlmostRelativeEquals(p2.b, p.b)


class CsvFileTextTests(amusetest.TestCase):

    def test1(self):
        print("Test 1: Read comma separated values (CSV) - specified attributes")
        contents = "#header\n1,2,3\n4,5,6\n7,8,9\n"
        data_stream = StringIO(contents)
        instance = text.CsvFileText(None, data_stream)
        instance.attribute_names = ['a', 'b', 'c']
        instance.attribute_types = [units.none, units.m, units.m/units.s]
        particles = instance.load()
        self.assertEqual(len(particles), 3)
        self.assertEqual(particles.a, [1, 4, 7] | units.none)
        self.assertEqual(particles.b, [2, 5, 8] | units.m)
        self.assertEqual(particles.c, [3, 6, 9] | units.m / units.s)

    def test2(self):
        print("Test 2: Read comma separated values (CSV) - attributes defined in header")
        contents = ("#a, b, c\n#no_system.get('none'),system.get('S.I.').base('length'),"
            "(system.get('S.I.').base('length') / system.get('S.I.').base('time'))\n"
            "#none, m, m/s\n1,2,3\n4,5,6\n7,8,9\n")
        data_stream = StringIO(contents)
        instance = text.CsvFileText(None, data_stream)
        particles = instance.load()
        self.assertEqual(len(particles), 3)
        self.assertEqual(particles.a, [1, 4, 7] | units.none)
        self.assertEqual(particles.b, [2, 5, 8] | units.m)
        self.assertEqual(particles.c, [3, 6, 9] | units.m / units.s)

    def test3(self):
        print("Test 3: Read comma separated values (CSV) - generic units")
        contents = ("#a,b,c\n"
            "#system.get('generic').base('mass'),system.get('generic').base('length'),"
            "(((system.get('generic').base('length')**2) * (system.get('generic').base('time')**-2)) * "
            "system.get('generic').base('mass'))\n"
            "#mass,length,length**2 * time**-2 * mass\n1.0,2.0,3.0\n4.0,5.0,6.0\n")
        data_stream = StringIO(contents)
        instance = text.CsvFileText(None, data_stream)
        particles = instance.load()
        self.assertEqual(len(particles), 2)
        self.assertEqual(particles.a, [1, 4] | generic_unit_system.mass)
        self.assertEqual(particles.b, [2, 5] | generic_unit_system.length)
        self.assertEqual(particles.c, [3, 6] | generic_unit_system.energy)

    def test4(self):
        print("Test 4: Write comma separated values (CSV) - specified attributes")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | units.none
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3

        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream, particles)
        instance.attribute_names = ['a', 'b']
        instance.attribute_types = [units.none, 100*units.cm]
        instance.store()

        contents = data_stream.getvalue()
        self.assertEqual("#a,b\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,100 * cm\n1.0,2.0\n4.0,5.0\n", contents)

    def test5(self):
        print("Test 5: Write comma separated values (CSV) - attributes defined automatically")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | units.none
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3

        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream, particles)
        instance.store()

        contents = data_stream.getvalue()
        self.assertEqual("#a,b,c\n"
            "#no_system.get('none'),system.get('S.I.').base('length'),"
            "((system.get('S.I.').base('length')**-3) * system.get('S.I.').base('mass'))\n"
            "#none,m,m**-3 * kg\n1.0,2.0,3.0\n4.0,5.0,6.0\n", contents)

    def test6(self):
        print("Test 6: Write comma separated values (CSV) - generic units")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | generic_unit_system.mass
        particles.b = [2, 5] | generic_unit_system.length
        particles.c = [3, 6] | generic_unit_system.energy
        particles.d = [4, 7] | generic_unit_system.temperature

        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream, particles)
        instance.store()

        contents = data_stream.getvalue()
        self.assertEqual("#a,b,c,d\n"
            "#system.get('generic').base('mass'),system.get('generic').base('length'),"
            "(((system.get('generic').base('length')**2) * (system.get('generic').base('time')**-2)) * "
            "system.get('generic').base('mass')),system.get('generic').base('thermodynamic temperature')\n"
            "#mass,length,length**2 * time**-2 * mass,thermodynamic temperature\n1.0,2.0,3.0,4.0\n4.0,5.0,6.0,7.0\n", contents)

    def test7(self):
        print("Test 7: Write CSV - quantities instead of set, names and types unspecified")
        a = [1.0, 4] | units.none
        b = [2.0, 5] | units.m

        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream)
        instance.quantities = [a, b]
        instance.store()
        self.assertEqual("#col(0),col(1)\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,m\n1.0,2.0\n4.0,5.0\n", data_stream.getvalue())

    def test8(self):
        print("Test 8: Write CSV - quantities instead of set, types unspecified")
        a = [1.0, 4] | units.none
        b = [2.0, 5] | units.m

        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream)
        instance.quantities = [a, b]
        instance.attribute_names = ['a', 'b']
        instance.store()

        self.assertEqual("#a,b\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,m\n1.0,2.0\n4.0,5.0\n", data_stream.getvalue())

    def test9(self):
        print("Test 9: Write CSV - quantities instead of set, names and types specified")
        a = [1.0, 4] | units.none
        b = [2.0, 5] | units.m

        data_stream = StringIO()
        instance = text.CsvFileText(None, data_stream)
        instance.quantities = [a, b]
        instance.attribute_names = ['a', 'b']
        instance.attribute_types = [units.none, 100*units.cm]
        instance.store()

        self.assertEqual("#a,b\n"
            "#no_system.get('none'),system.get('S.I.').base('length')\n"
            "#none,100 * cm\n1.0,2.0\n4.0,5.0\n", data_stream.getvalue())

    def test10(self):
        print("Test 10: User interface (write_set_to_file and read_set_from_file)")
        particles = datamodel.Particles(2)
        particles.a = [1, 4] | units.none
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3
        io.write_set_to_file(particles, "test_textio.csv", "csv")

        read_particles = io.read_set_from_file("test_textio.csv", format="csv")
        self.assertEqual(len(read_particles), 2)
        self.assertEqual(read_particles.a, [1, 4] | units.none)
        self.assertEqual(read_particles.b, [2, 5] | units.m)
        self.assertEqual(read_particles.c, [3, 6] | units.kg / units.m**3)
        os.remove("test_textio.csv")

    def test11(self):
        particles = datamodel.Particles(2)
        particles.a = [1, 4]
        particles.b = [2, 5] | units.m
        particles.c = [3, 6] | units.kg / units.m**3
        io.write_set_to_file(
            particles,
            "test_textio.csv",
            "csv",
            attribute_type=(None, units.kg / units.m**3, units.m),
            attribute_name=('a', 'c', 'b')
        )

        read_particles = io.read_set_from_file(
            "test_textio.csv",
            format="csv"
        )
        self.assertEqual(len(read_particles), 2)
        self.assertEqual(read_particles.a, [1, 4])
        self.assertEqual(read_particles.b, [2, 5] | units.m)
        self.assertEqual(read_particles.c, [3, 6] | units.kg / units.m**3)
        os.remove("test_textio.csv")


class Athena3DTextTests(amusetest.TestCase):

    def test1(self):
        contents = """\
        # Nx1 = 128
        # x1-size = 1
        # Time = 0.103125
        #
        # [1]=i-zone [2]=x1 [3]=d [4]=M1 [5]=M2 [6]=M3 [7]=P [8]=E [9]=B1c [10]=B2c [11]=B3c [12]=B1i [13]=B1i [14]=B1i
        #
          4  3.90625e-03  1.00000e+00 -8.66511e-07  4.08477e-07  1.44419e-07  6.00000e-01  2.52500e+00  1.00000e+00  1.41421e+00  5.00000e-01  1.00000e+00  1.41421e+00  5.00000e-01
        """
        data_file = StringIO(textwrap.dedent(contents))
        instance = text.Athena3DText("test.tab", data_file)
        particles = instance.load()

        self.assertEqual(len(particles), 1)
        self.assertEqual(particles[0].x1, 3.90625e-03 | units.none)
        self.assertEqual(particles[0].i_zone, 4 | units.none)
        self.assertEqual(particles[0].M1, -8.66511e-07 | units.none)



from amuse.test import amusetest

from amuse.datamodel import Particles
from amuse.units import units
from amuse.community.seba import Seba
from amuse.community.bse import Bse
from amuse.support.console import set_printing_strategy
from amuse.units import units


class TestsForIssue850(amusetest.TestCase):
    def create_stars_and_binaries(self, binary_pair=[0, 1]):
        stars = Particles(3)
        stars[0].initial_mass = 14 | units.MSun
        stars[1].initial_mass = 10 | units.MSun
        stars[2].initial_mass = 9 | units.MSun
        stars.mass = stars.initial_mass

        binary = Particles(1)
        binary.semi_major_axis = 30000 | units.RSun
        binary.eccentricity = 0.3
        binary.child1 = stars[binary_pair[0]]
        binary.child2 = stars[binary_pair[1]]

        return stars, binary

    def test_do_all_stars_evolve_no_binary(self, code=Seba):
        "Test if all stars evolve (no binaries)"
        stars, binary = self.create_stars_and_binaries(binary_pair=[0, 1])

        instance = code()
        instance.particles.add_particles(stars)

        channel_stars = instance.particles.new_channel_to(stars)
        channel_stars.copy()

        end_time = 13500. | units.Myr

        instance.evolve_model(end_time)
        channel_stars.copy()

        # All stars need to have evolved
        print(
            f"t: {end_time}, m: {stars[0].mass} {stars[1].mass} {stars[2].mass} "
        )
        for i in range(len(stars)):
            self.assertNotEqual(stars[i].mass, stars[i].initial_mass)
        instance.stop()

    def _test_do_all_stars_evolve_no_binary_bse(self):
        self.test_do_all_stars_evolve_no_binary(code=Bse)

    def test_does_seba_evolve_stars_not_in_binary(self, code=Seba):
        """
        Tests if code evolves any star that is not in a binary if a binary is
        added.
        """
        set_printing_strategy('default')
        stars, binary = self.create_stars_and_binaries(binary_pair=[0, 1])
        additional_stars = Particles(5)
        additional_stars.original_mass = [11, 20, 4, 2, 1] | units.MSun
        additional_stars.mass = additional_stars.original_mass
        stars.add_particles(additional_stars)

        instance = code()
        instance.particles.add_particles(stars)
        instance.binaries.add_particles(binary)

        channel_stars = instance.particles.new_channel_to(stars)
        channel_binary = instance.binaries.new_channel_to(binary)
        channel_stars.copy()
        channel_binary.copy()

        end_time = 13500. | units.Myr

        instance.evolve_model(end_time)
        channel_stars.copy()
        channel_binary.copy()

        # All stars need to have evolved
        print(
            f"t: {end_time}, m: {stars[0].mass} {stars[1].mass} {stars[2].mass} "
        )
        for i in range(len(stars)):
            self.assertNotEqual(stars[i].mass, stars[i].initial_mass)
        instance.stop()

    def _test_does_bse_evolve_stars_not_in_binary(self):
        self.test_does_seba_evolve_stars_not_in_binary(code=Bse)

    def test_do_all_stars_evolve_default_binary(self):
        "Test if all stars evolve (default order binary)"
        stars, binary = self.create_stars_and_binaries(binary_pair=[0, 1])

        instance = Seba()
        instance.particles.add_particles(stars)
        instance.binaries.add_particles(binary)

        channel_stars = instance.particles.new_channel_to(stars)
        channel_binary = instance.binaries.new_channel_to(binary)
        channel_stars.copy()
        channel_binary.copy()

        end_time = 13500. | units.Myr

        instance.evolve_model(end_time)
        channel_stars.copy()
        channel_binary.copy()

        # All stars need to have evolved
        print(
            f"t: {end_time}, m: {stars[0].mass} {stars[1].mass} {stars[2].mass} "
        )
        for i in range(len(stars)):
            self.assertNotEqual(stars[i].mass, stars[i].initial_mass)
        instance.stop()

    def test_do_all_stars_evolve_alternative_binary(self):
        "Test if all stars evolve (other stars in the binary)"
        stars, binary = self.create_stars_and_binaries(binary_pair=[0, 2])

        instance = Seba()
        instance.particles.add_particles(stars)
        instance.binaries.add_particles(binary)

        channel_stars = instance.particles.new_channel_to(stars)
        channel_binary = instance.binaries.new_channel_to(binary)
        channel_stars.copy()
        channel_binary.copy()

        end_time = 13500. | units.Myr

        instance.evolve_model(end_time)
        channel_stars.copy()
        channel_binary.copy()

        # All stars need to have evolved
        print(
            f"t: {end_time}, m: {stars[0].mass} {stars[1].mass} {stars[2].mass} "
        )
        for i in range(len(stars)):
            self.assertNotEqual(stars[i].mass, stars[i].initial_mass)
        instance.stop()

    def test_do_all_stars_evolve_alternative_binary_extra_stars(self):
        "Test if all stars evolve (other stars in the binary, extra stars)"
        stars, binary = self.create_stars_and_binaries(binary_pair=[0, 2])
        additional_stars = Particles(5)
        additional_stars.original_mass = [11, 20, 4, 2, 1] | units.MSun
        additional_stars.mass = additional_stars.original_mass
        stars.add_particles(additional_stars)

        instance = Seba()
        instance.particles.add_particles(stars)
        instance.binaries.add_particles(binary)

        channel_stars = instance.particles.new_channel_to(stars)
        channel_binary = instance.binaries.new_channel_to(binary)
        channel_stars.copy()
        channel_binary.copy()

        end_time = 13500. | units.Myr

        instance.evolve_model(end_time)
        channel_stars.copy()
        channel_binary.copy()

        # All stars need to have evolved
        print(
            f"t: {end_time}, m: {stars[0].mass} {stars[1].mass} {stars[2].mass} "
        )
        for i in range(len(stars)):
            self.assertNotEqual(stars[i].mass, stars[i].initial_mass)
        instance.stop()


if __name__ == "__main__":
    x = TestsForIssue850()
    x.test_do_all_stars_evolve_no_binary()
    x.test_does_seba_evolve_stars_not_in_binary()
    x.test_do_all_stars_evolve_default_binary()
    x.test_do_all_stars_evolve_alternative_binary_extra_stars()
    x.test_do_all_stars_evolve_alternative_binary()



from amuse.test import amusetest

import numpy
import sys


from amuse.support.exceptions import AmuseException
from amuse.units.quantities import *
from amuse.units import si
from amuse.units import units
from amuse.units import trigo
from amuse.units import nbody_system
from amuse.units import quantities
from amuse.units import core


from amuse import datamodel


class TestQuantities(amusetest.TestCase):

    def test1(self):
        x = 1.0 | si.kg
        self.assertTrue(isinstance(x, ScalarQuantity))
        x = [1.0, 2.0, 3.0] | si.kg
        self.assertTrue(isinstance(x, VectorQuantity))

    def test2(self):
        x = [1.0, 2.0, 3.0] | si.kg
        y = [2.0, 3.0, 4.0] | si.kg
        xy = x * y
        self.assertTrue(isinstance(xy, VectorQuantity))

    def test3(self):
        x = [1.0, 2.0, 3.0] | si.kg
        y = [2.0, 3.0, 4.0] | si.kg
        self.assertTrue(isinstance(x[0], ScalarQuantity))
        self.assertEqual(str(x[1]), "2.0 kg")

    def test4(self):
        g = si.kg / 1000
        x = [1.0, 2.0, 3.0] | si.kg
        self.assertEqual(str(x), "[1.0, 2.0, 3.0] kg")
        x[0] = 3000.0 | g
        self.assertEqual(str(x), "[3.0, 2.0, 3.0] kg")

    def test5(self):
        number_of_stars = 10
        stars = datamodel.Particles(number_of_stars)
        stars.position = [0, 0, 0] | units.km
        for i, star in enumerate(stars):
            star.position = units.km.new_quantity([float(i+1), float((i+1)*2), float(-1 * (i+1))])

        minpos = [float(sys.maxsize)] * 3 | units.m
        maxpos = [-float(sys.maxsize)] * 3 | units.m
        for star in stars:
            for i in range(3):
                if star.position[i] < minpos[i]:
                    minpos[i] = star.position[i]
                if star.position[i] > maxpos[i]:
                    maxpos[i] = star.position[i]

        self.assertEqual(str(minpos), "[1000.0, 2000.0, -10000.0] m")
        self.assertEqual(str(maxpos), "[10000.0, 20000.0, -1000.0] m")

    def test6(self):
        x = [1.0, 2.0, 3.0] | si.kg
        y = x.copy()
        y[0] = 3.0 | si.kg
        self.assertEqual(x[0].value_in(si.kg), 1.0)
        self.assertEqual(y[0].value_in(si.kg), 3.0)

    def test7(self):
        x = 2.0 | si.kg
        y = 1 / x
        self.assertEqual(y.value_in(1/si.kg), 0.5)

    def test8(self):
        x = (1.0, 2.0, 3.0) | si.kg
        self.assertTrue(isinstance(x, VectorQuantity))

    def test9(self):
        converter = nbody_system.nbody_to_si(1 | si.kg, 2 | si.s)
        self.assertEqual(0.0 | nbody_system.mass, converter.to_nbody(zero))
        self.assertEqual(converter.to_nbody(zero), 0.0 | nbody_system.mass)

    def test10(self):
        self.assertEqual(1 | units.m, 1 | units.m)
        self.assertTrue(1 | units.m == 1 | units.m)
        self.assertFalse(1 | units.m == 2 | units.m)
        self.assertTrue(1 | units.m != 2 | units.m)
        self.assertFalse(1 | units.m != 1 | units.m)
        self.assertTrue(1 | units.m >= 1 | units.m)
        self.assertFalse(1 | units.m >= 2 | units.m)
        self.assertTrue(1 | units.m <= 1 | units.m)
        self.assertFalse(1 | units.m <= 0 | units.m)
        self.assertTrue(1 | units.m > 0 | units.m)
        self.assertFalse(1 | units.m > 1 | units.m)
        self.assertTrue(1 | units.m < 3 | units.m)
        self.assertFalse(1 | units.m < 0 | units.m)

    def test11(self):
        self.assertEqual([1] | units.m, [1] | units.m)
        self.assertTrue([1] | units.m == [1] | units.m)
        self.assertFalse([1] | units.m == [2] | units.m)
        self.assertTrue([1] | units.m != [2] | units.m)
        self.assertFalse([1] | units.m != [1] | units.m)
        self.assertTrue([1] | units.m >= [1] | units.m)
        self.assertFalse([1] | units.m >= [2] | units.m)
        self.assertTrue([1] | units.m <= [1] | units.m)
        self.assertFalse([1] | units.m <= [0] | units.m)
        self.assertTrue([1] | units.m > [0] | units.m)
        self.assertFalse([1] | units.m > [1] | units.m)
        self.assertTrue([1] | units.m < [3] | units.m)
        self.assertFalse([1] | units.m < [0] | units.m)

    def test12(self):
        self.assertEqual(zero, zero)
        self.assertTrue(zero == zero)
        self.assertFalse(zero == zero + (1 | units.m))
        self.assertFalse(zero + (1 | units.m) == zero)
        self.assertTrue(zero != zero + (1 | units.m))
        self.assertFalse(zero != zero)
        self.assertTrue(zero >= zero)
        self.assertFalse(zero >= zero + (1 | units.m))
        self.assertTrue(zero <= zero + (1 | units.m))
        self.assertFalse(zero <= zero - (1 | units.m))
        self.assertTrue(zero > zero - (1 | units.m))
        self.assertFalse(zero > zero)
        self.assertTrue(zero < zero + (1 | units.m))
        self.assertFalse(zero < zero - (1 | units.m))
        self.assertTrue(zero == 0 | units.m)

    def test13(self):
        self.assertEqual('a', 'a')
        self.assertTrue('a' == 'a')
        self.assertFalse('a' == 'ab')
        self.assertTrue('a' != 'A')
        self.assertFalse('a' != 'a')
        self.assertTrue('b' >= 'a')
        self.assertFalse('B' >= 'a')
        self.assertTrue('a' <= 'ab')
        self.assertFalse('a' <= 'A')
        self.assertTrue('a' > 'A')
        self.assertFalse('a' > 'a')
        self.assertTrue('a' < 'b')
        self.assertFalse('a' < 'B')

    def test14(self):
        # Tests for 'is_quantity'
        self.assertTrue(is_quantity(0 | units.kg))
        self.assertTrue(is_quantity(1 | units.none))
        self.assertTrue(is_quantity([1.0, 2.0, 3.0] | units.m))
        self.assertFalse(is_quantity(1))
        self.assertFalse(is_quantity(1.0))
        self.assertFalse(is_quantity("string"))

    def test15(self):
        # Tests for 'to_quantity'
        self.assertTrue(is_quantity(to_quantity(0 | units.kg)))
        self.assertTrue(is_quantity(to_quantity(1 | units.none)))
        self.assertTrue(is_quantity(to_quantity([1.0, 2.0, 3.0] | units.m)))
        self.assertTrue(is_quantity(to_quantity(1)))
        self.assertTrue(is_quantity(to_quantity(1.0)))
        masses = [1, 2, 3] | units.kg
        self.assertTrue(to_quantity(masses) is masses)
        numbers = [1, 2, 3]
        self.assertFalse(to_quantity(numbers) is numbers | units.none)
        self.assertTrue(numpy.all(to_quantity(numbers) == numbers | units.none))

    def test16(self):
        # Tests for add/sub of quantity (with none unit) and number
        self.assertEqual((2.0 | units.none) + 1.0,  3.0)
        self.assertEqual((2.0 | units.none) - 1.0,  1.0)
        self.assertEqual(1.0 + (2.0 | units.none),  3.0)
        self.assertEqual(1.0 - (2.0 | units.none), -1.0)

    def test17(self):
        # Tests for add/sub of quantity (with other unit) and number
        number = 1.0
        quantity = 2.0 | units.m
        self.assertTrue(number.__add__(quantity) is NotImplemented)
        self.assertRaises(AmuseException, quantity.__radd__, number)
        self.assertTrue(number.__radd__(quantity) is NotImplemented)
        self.assertRaises(AmuseException, quantity.__add__, number)
        self.assertTrue(number.__sub__(quantity) is NotImplemented)
        self.assertRaises(AmuseException, quantity.__rsub__, number)
        self.assertTrue(number.__rsub__(quantity) is NotImplemented)
        self.assertRaises(AmuseException, quantity.__sub__, number)
        # in other words...
        self.assertRaises(AmuseException, lambda: number + quantity,
            expected_message="Cannot express none in m, the units do not have the same bases")
        self.assertRaises(AmuseException, lambda: quantity + number,
            expected_message="Cannot express none in m, the units do not have the same bases")
        self.assertRaises(AmuseException, lambda: number - quantity,
            expected_message="Cannot express none in m, the units do not have the same bases")
        self.assertRaises(AmuseException, lambda: quantity - number,
            expected_message="Cannot express none in m, the units do not have the same bases")

    def test18(self):
        quantity = 'string'
        self.assertEqual(quantity,  'string')
        quantity = 'string'
        self.assertEqual(quantity, 'string')

    def test19(self):
        x = 1.0 | si.kg
        self.assertTrue(x == x.amin())
        self.assertTrue(x == x.prod())
        self.assertTrue(x == x.sorted())
        self.assertTrue(x == x.amax())
        self.assertTrue(x == x.sum())

    def test20(self):
        lengths = [] | units.m
        lengths.append(1 | units.m)
        self.assertEqual(lengths, [1] | units.m)
        lengths.append(2 | units.m)
        self.assertEqual(lengths, [1, 2] | units.m)

        positions = [] | units.m
        positions.append([1, 2, 3] | units.m)
        self.assertEqual(positions, [[1, 2, 3]] | units.m)
        positions.append([4, 5, 6] | units.m)
        self.assertEqual(positions, [[1, 2, 3], [4, 5, 6]] | units.m)

        two_positions = [] | units.m
        two_positions.append([[1, 2, 3], [-1, -2, -3]] | units.m)
        self.assertEqual(two_positions, [[[1, 2, 3], [-1, -2, -3]]] | units.m)
        two_positions.append([[4, 5, 6], [7, 8, 9]] | units.m)
        self.assertEqual(two_positions, [[[1, 2, 3], [-1, -2, -3]], [[4, 5, 6], [7, 8, 9]]] | units.m)
        # Appending quantities with incompatible shapes:
        two_positions.append(99 | units.m)
        self.assertEqual(two_positions, [1, 2, 3, -1, -2, -3, 4, 5, 6, 7, 8, 9, 99] | units.m)

    def test21(self):
        zero_vector = zero.as_vector_with_length(3)
        self.assertEqual(str(zero_vector), "[0.0, 0.0, 0.0] zero")

        self.assertEqual(zero_vector + (1 | units.m), [1, 1, 1] | units.m)
        self.assertEqual(zero_vector - (1 | units.m), [-1, -1, -1] | units.m)
        self.assertEqual((1 | units.m) + zero_vector, [1, 1, 1] | units.m)
        self.assertEqual((1 | units.m) - zero_vector, [1, 1, 1] | units.m)

        self.assertEqual(zero_vector + ([1, 1, 1] | units.m), [1, 1, 1] | units.m)
        self.assertEqual(zero_vector - ([1, 1, 1] | units.m), [-1, -1, -1] | units.m)
        self.assertEqual(([1, 1, 1] | units.m) + zero_vector, [1, 1, 1] | units.m)
        self.assertEqual(([1, 1, 1] | units.m) - zero_vector, [1, 1, 1] | units.m)

        for one_zero in zero_vector:
            self.assertEqual(one_zero, zero)

        self.assertEqual(zero_vector[0:2], zero.as_vector_with_length(2))

    def test22(self):
        x = numpy.asarray([1, 2, 3, 4])
        y = 2 | units.m
        self.assertTrue(is_quantity(y * x))
        self.assertAlmostRelativeEquals(y*x, [2, 4, 6, 8] | units.m)
        self.assertTrue(is_quantity(x * y))
        self.assertAlmostRelativeEquals(x*y, [2, 4, 6, 8] | units.m)

    def test23(self):
        z = zero.as_vector_with_length(2)
        self.assertEqual(len(z), 2)
        z += 1 | units.kg
        self.assertEqual(z.unit, units.kg)
        self.assertEqual(z, [1, 1] | units.kg)

    def xtest24(self):
        rhs = 2 | units.MSun / units.AU
        lhs = 3 | units.AU
        product = rhs * lhs
        product_unit = product.unit
        print(product)
        self.assertTrue(product_unit is units.MSun)

    def xtest25(self):
        rhs = 2.0 | (2 * units.MSun)**2 / units.AU
        lhs = 3.0 | units.MSun
        product = rhs / lhs
        product_unit = product.unit
        print(product)
        self.assertEqual(product_unit, units.MSun / units.AU)
        self.assertEqual(product_unit.local_factor, 4)

    def xtest26(self):
        rhs = 2.0 | units.AU / (2 * units.MSun)**2
        lhs = 3.0 | units.MSun
        product = rhs * lhs
        product_unit = product.unit
        print(product)
        print(type(product_unit))
        self.assertEqual(product_unit, units.AU / units.MSun)
        self.assertEqual(product_unit.local_factor, 1/4.0)

    def test27(self):
        a = [1. | units.kg, 2. | units.kg, 3000. | units.g, 4. | (1000*units.g)]
        b = VectorQuantity.new_from_scalar_quantities(*a)
        c = [1., 2., 3., 4.] | units.kg
        print(a[0].unit == a[2].unit)
        self.assertEqual(b, c)

    def test28(self):
        a = [1. | units.kg, 2. | units.kg, 3000. | units.m, 4. | (1000*units.g)]
        try:
            b = VectorQuantity.new_from_scalar_quantities(*a)
            raise Exception("expect error")
        except:
            pass

    def test29(self):
        one_inch = 2.54 | units.cm
        self.assertFalse(isinstance(one_inch, core.unit))
        self.assertTrue(isinstance(one_inch.as_unit(), core.unit))
        self.assertEqual(one_inch.as_unit(), 2.54 * units.cm)
        self.assertEqual(1 | one_inch.as_unit(), 2.54 | units.cm)

    def test30(self):
        a = 1.5 | units.km
        b = 1000. | units.m
        self.assertEqual(a % b, 500. | units.m)
        a = [1.5, 1.75] | units.km
        b = 1000. | units.m
        self.assertEqual(a % b, [500., 750] | units.m)
        a = [1.5, 1.75] | units.km
        b = [1000., 500.] | units.m
        self.assertEqual(a % b, [500., 250.] | units.m)
        a = [1.5] | units.km
        b = [1000., 500.] | units.m
        self.assertEqual(a % b, [500., 0.] | units.m)

    def test31(self):
        """ 
        test trigonometric unit stuff
        """
        self.assertEqual(units.pi, numpy.pi)
        a = units.pi
        self.assertEqual(trigo.to_rad(a), numpy.pi | units.rad)
        self.assertEqual(trigo.to_deg(a), 180. | units.deg)
        self.assertEqual(trigo.to_rev(a), 0.5 | units.rev)
        a = 90 | units.deg
        self.assertEqual(trigo.to_rad(a), numpy.pi/2 | units.rad)
        self.assertEqual(trigo.to_deg(a), 90. | units.deg)
        self.assertEqual(trigo.to_rev(a), 0.25 | units.rev)
        a = 0.75 | units.rev
        self.assertEqual(trigo.to_rad(a), 3/2.*numpy.pi | units.rad)
        self.assertEqual(trigo.to_deg(a), 270. | units.deg)
        self.assertEqual(trigo.to_rev(a), 0.75 | units.rev)
        a = 2*numpy.pi
        self.assertEqual(trigo.to_rad(a), 2*numpy.pi | units.rad)
        self.assertEqual(trigo.to_deg(a), 360. | units.deg)
        self.assertEqual(trigo.to_rev(a), 1. | units.rev)

        a = 45. | units.deg
        self.assertEqual(trigo.sin(a), numpy.sin(45./180*numpy.pi))
        self.assertEqual(trigo.cos(a), numpy.cos(45./180*numpy.pi))
        self.assertEqual(trigo.tan(a), numpy.tan(45./180*numpy.pi))

        a = 1. | units.rad
        self.assertEqual(trigo.sin(a), numpy.sin(1.))
        self.assertEqual(trigo.cos(a), numpy.cos(1.))
        self.assertEqual(trigo.tan(a), numpy.tan(1.))

        a = 0.125 | units.rev
        self.assertEqual(trigo.sin(a), numpy.sin(45./180*numpy.pi))
        self.assertEqual(trigo.cos(a), numpy.cos(45./180*numpy.pi))
        self.assertEqual(trigo.tan(a), numpy.tan(45./180*numpy.pi))

        a = 45. | units.deg
        self.assertAlmostEqual(trigo.arcsin(trigo.sin(a)), 45. | units.deg, 13)
        self.assertAlmostEqual(trigo.arccos(trigo.cos(a)), 45. | units.deg, 13)
        self.assertAlmostEqual(trigo.arctan(trigo.tan(a)), 45. | units.deg, 13)

    def test32(self):
        a = numpy.array([[1., 2., 3.], [4., 5., 6.]]) | units.m
        b = numpy.array([[1., 2., 3.], [4., 5., 6.]])
        self.assertEqual(list(a.flatten()), list(a.flat))
        flat1 = b.flat
        flat2 = a.flat
        self.assertEqual(flat1[2:5], flat2[2:5].number)
        next(flat1)
        next(flat2)
        self.assertEqual(flat1.index, flat2.index)
        self.assertEqual(flat1.base, flat2.base.number)
        self.assertEqual(flat1.copy(), flat2.copy().number)

    def test32b(self):
        a = numpy.array([[1., 2., 3.], [4., 5., 6.]]) | units.m
        b = numpy.array([[1., 2., 3.], [4., 5., 6.]])
        flat1 = b.flat
        flat2 = a.flat
        self.assertEqual(flat1[2:5], flat2[2:5].number)
        self.assertEqual(flat1, flat2.number)

        flat2[:] = numpy.arange(6) | units.cm
        a_ = numpy.array([[0., 1., 2.], [3., 4., 5.]]) | units.cm
        self.assertEqual(a, a_)

    def test33(self):
        a = [1, 2, 3, 4]
        b = new_quantity_nonone(a, units.none)
        self.assertEqual(len(a), len(b))
        b = new_quantity_nonone(a, 2*units.none)
        self.assertEqual(len(a), len(b))

    def test34(self):
        a = [1, 2, 3, 4, 5] | units.m
        x = a.value_in(units.cm)
        x[0] = -1
        self.assertEqual(a, [1, 2, 3, 4, 5] | units.m)

        a = [1, 2, 3, 4, 5] | units.m
        x = a.value_in(units.m)
        x[0] = -1
        self.assertEqual(a, [1, 2, 3, 4, 5] | units.m)

    def test_mean(self):
        x = [1.0, 2.0, 3.0] | si.kg
        mask = [True, True, False]
        self.assertEqual(x.mean(), 2.0 | si.kg)
        self.assertEqual(x.mean(where=mask), 1.5 | si.kg)


class TestAdaptingVectorQuantities(amusetest.TestCase):

    def test1(self):
        x = AdaptingVectorQuantity()
        self.assertEqual(x.append.__name__, "append_start")
        x.append(1 | units.kg)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 1)
        self.assertEqual(x.append.__name__, "append_normal")

        self.assertTrue(isinstance(x._number_list, list))
        self.assertFalse(isinstance(x._number_list, numpy.ndarray))
        self.assertTrue(isinstance(x.number, numpy.ndarray))
        self.assertFalse(isinstance(x.number, list))
        self.assertEqual(x._number_list, [1])
        self.assertEqual(x.number, numpy.array([1]))

    def test2(self):
        x = AdaptingVectorQuantity()
        self.assertEqual(len(x), 0)
        self.assertEqual(len(x.number), 0)
        self.assertEqual(str(x), '[]')
        x.append(1 | units.kg)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 1)
        self.assertEqual(str(x), '[1] kg')
        x.append(2 | units.kg)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 2)
        self.assertEqual(str(x), '[1, 2] kg')

    def test3(self):
        x = AdaptingVectorQuantity()
        x.extend([1, 2, 3] | units.kg)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 3)
        self.assertEqual(x.number, numpy.array([1, 2, 3]))

        x.extend([1, 2, 3] | units.g)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 6)
        self.assertAlmostRelativeEquals(x, [1000, 2000, 3000, 1, 2, 3] | units.g)

    def test4(self):
        x = AdaptingVectorQuantity()
        x.prepend(1 | units.kg)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 1)
        self.assertEqual(str(x), '[1] kg')
        x.prepend(2 | units.kg)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 2)
        self.assertEqual(str(x), '[2, 1] kg')

    def test5(self):
        # Everything mixed...
        x = AdaptingVectorQuantity()
        x.extend([3, 4, 5] | units.kg)
        x.append(6 | units.kg)
        x.prepend(2 | units.kg)
        x.extend([7000, 8000, 9000] | units.g)
        x.prepend(1000 | units.g)
        x.append(10000 | units.g)
        self.assertEqual(x.unit, units.kg)
        self.assertEqual(len(x), 10)
        self.assertEqual(x.number, numpy.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]))
        self.assertEqual(x, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] | units.kg)

    def test6(self):
        x = 6 | units.kg
        y = 5 | units.kg
        self.assertEqual(x/y, 6/5)
        self.assertEqual(x//y, 6//5)
        self.assertEqual(operator.__truediv__(x, y), 1.2)


class TestNumpyFunctionWithUnits(amusetest.TestCase):

    def test1(self):
        array = quantities.arange(0 | units.kg, 10 | units.kg, 1 | units.kg)
        self.assertEqual(len(array),  10)
        self.assertAlmostRelativeEquals(array, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9] | units.kg)

    def test2(self):
        array = quantities.linspace(0 | units.kg, 10 | units.kg, 11)
        self.assertEqual(len(array),  11)
        self.assertAlmostRelativeEquals(array, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10] | units.kg)

    def test3(self):
        array = quantities.linspace(0, 10, 11)
        self.assertEqual(len(array),  11)
        self.assertAlmostRelativeEquals(array, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    def test4(self):
        x = quantities.arange(0 | units.yr, 10 | units.yr, 1 | units.yr)
        y = (2.0 | units.km) * (x / (2.0 | units.yr))**2 + (20.0 | units.km)

        fit = quantities.polyfit(x, y, 2)

        self.assertEqual(len(fit), 3)
        self.assertEqual(fit[0].unit, units.km/units.yr**2)

        fit_values = quantities.polyval(fit, x)

        self.assertEqual(fit_values.shape, x.shape)
        self.assertEqual(y.unit, fit_values.unit)

        self.assertAlmostRelativeEquals(y, fit_values, 1)

    def test5(self):
        a = [1, 2, 3] | units.m
        b = [4, 5, 6] | units.m

        ab1 = quantities.column_stack((a, b))
        ab2 = quantities.column_stack((a.number, b.number)) | units.m

        self.assertEqual(ab1, ab2)



import os
import sys
import numpy as np
from amuse.test.amusetest import TestWithMPI

from amuse.community.mercury.interface import MercuryInterface, MercuryWayWard, Mercury
from amuse.ext.solarsystem import new_solar_system_for_mercury, new_solar_system
from amuse.units import nbody_system
from amuse.units import units, constants
from amuse import datamodel
from amuse.ic import plummer
DUMMYID = 0

try:
    from matplotlib import pyplot
    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False


class TestMercuryInterface(TestWithMPI):

    def setUp(self):
        super(TestWithMPI, self).setUp()

    def test1(self):
        instance = MercuryInterface()
        instance.stop()

    def test2(self):
        instance = MercuryInterface()
        instance.initialize_code()
        instance.stop()

    def test3(self):
        instance = MercuryInterface()
        instance.initialize_code()
        instance.cleanup_code()
        instance.stop()

    def test4(self):
        instance = MercuryInterface()
        instance.initialize_code()
        mass, err = instance.get_central_mass(DUMMYID)
        self.assertEqual(mass, 1.0)
        radius, err = instance.get_central_radius(DUMMYID)
        self.assertEqual(radius, .005)
        j2, j4, j6, err = instance.get_central_oblateness(DUMMYID)
        self.assertEqual(j2, 0.0)
        self.assertEqual(j4, 0.0)
        self.assertEqual(j6, 0.0)
        lx, ly, lz, err = instance.get_central_spin(DUMMYID)
        self.assertEqual(lx, 0.0)
        self.assertEqual(ly, 0.0)
        self.assertEqual(lz, 0.0)
        instance.stop()

    def test5(self):
        instance = MercuryInterface()
        instance.initialize_code()
        instance.set_central_mass(DUMMYID, 2.0)
        instance.set_central_radius(DUMMYID, 0.0625)
        instance.set_central_oblateness(DUMMYID, 0.001, 0.002, 0.003)
        instance.set_central_spin(DUMMYID, -0.1, -0.2, -0.3)

        mass, err = instance.get_central_mass(DUMMYID)
        self.assertEqual(mass, 2.0)
        radius, err = instance.get_central_radius(DUMMYID)
        self.assertEqual(radius, .0625)
        j2, j4, j6, err = instance.get_central_oblateness(DUMMYID)
        self.assertEqual(j2, 0.001)
        self.assertEqual(j4, 0.002)
        self.assertEqual(j6, 0.003)
        lx, ly, lz, err = instance.get_central_spin(DUMMYID)
        self.assertEqual(lx, -0.1)
        self.assertEqual(ly, -0.2)
        self.assertEqual(lz, -0.3)
        instance.stop()

    def test6(self):
        instance = MercuryInterface()
        instance.initialize_code()
        instance.commit_particles()
        instance.cleanup_code()
        instance.stop()

    def test7(self):
        instance = MercuryInterface()
        instance.initialize_code()
        time, err = instance.get_time()
        self.assertEqual(time, 0.0)
        instance.stop()

    def test8(self):
        instance = MercuryInterface()
        instance.initialize_code()
        n, err = instance.get_number_of_orbiters()
        self.assertEqual(n, 0)
        instance.stop()

    def test9(self):
        instance = MercuryInterface()
        instance.initialize_code()
        pid, err = instance.new_orbiter(0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.)
        n, err = instance.get_number_of_orbiters()
        self.assertEqual(n, 1)
        instance.stop()

    def test10(self):
        instance = MercuryInterface()
        instance.initialize_code()
        pid, err = instance.new_orbiter(0.5, 1., 2., 3., 4., 5., 6., 7., 1., 2., 3., 8.)
        mass, dens, x, y, z, vx, vy, vz, sx, sy, sz, celimit, err = instance.get_orbiter_state(pid)
        self.assertEqual(mass, 0.5)
        self.assertEqual(dens, 1.)
        self.assertEqual(x, 2.)
        self.assertEqual(y, 3.)
        self.assertEqual(z, 4.)
        self.assertEqual(vx, 5.)
        self.assertEqual(vy, 6.)
        self.assertEqual(vz, 7.)
        self.assertEqual(sx, 1.)
        self.assertEqual(sy, 2.)
        self.assertEqual(sz, 3.)
        self.assertEqual(celimit, 8.)
        instance.stop()

    def test11(self):
        instance = MercuryInterface()
        instance.initialize_code()
        mass = 3.04043264264672381E-06
        dens = 5.52
        x = 2.42093942183383037E-01
        y = -9.87467766698604366E-01
        z = -4.54276292555233496E-06
        vx = 1.64294055023289365E-02
        vy = 4.03200725816140870E-03
        vz = 1.13609607260006795E-08
        sx = sy = sz = 0.
        celimit = 20.
        pid, err = instance.new_orbiter(mass, dens, x, y, z, vx, vy, vz, sx, sy, sz, celimit)
        instance.commit_particles()
        de, err = instance.get_energy_deviation()
        self.assertEqual(de, 0.0)
        etot, err = instance.get_total_energy()
        ep, err = instance.get_potential_energy()
        ek, err = instance.get_kinetic_energy()
        self.assertAlmostEqual(etot, ep+ek, 15)
        instance.stop()

    def test12(self):
        instance = MercuryInterface()
        instance.initialize_code()
        mass = 3.04043264264672381E-06
        dens = 5.52
        x = 2.42093942183383037E-01
        y = -9.87467766698604366E-01
        z = -4.54276292555233496E-06
        vx = 1.64294055023289365E-02
        vy = 4.03200725816140870E-03
        vz = 1.13609607260006795E-08
        sx = sy = sz = 0.
        celimit = 20.
        pid, err = instance.new_orbiter(mass, dens, x, y, z, vx, vy, vz, sx, sy, sz, celimit)
        instance.commit_particles()
        etot1, err = instance.get_total_energy()
        err = instance.evolve_model(8*100000.)
        etot2, err = instance.get_total_energy()
        self.assertAlmostEqual(etot1, etot2, 7)
        instance.stop()

    def test13(self):
        instance = MercuryInterface()
        instance.initialize_code()
        mass = 3.04043264264672381E-06
        dens = 5.52
        x = 2.42093942183383037E-01
        y = -9.87467766698604366E-01
        z = -4.54276292555233496E-06
        vx = 1.64294055023289365E-02
        vy = 4.03200725816140870E-03
        vz = 1.13609607260006795E-08
        sx = sy = sz = 0.
        celimit = 20.
        pid, err = instance.new_orbiter(mass, dens, x, y, z, vx, vy, vz, sx, sy, sz, celimit)
        instance.commit_particles()
        instance.enable_stopping_condition(3)
        instance.set_stopping_condition_timeout_parameter(0.01)
        etot1, err = instance.get_total_energy()
        err = instance.evolve_model(8*1e100)
        self.assertTrue(instance.is_stopping_condition_set(3)['result'] == 1)
        etot2, err = instance.get_total_energy()
        self.assertAlmostEqual(etot1, etot2, 7)
        instance.stop()

    def test14(self):
        instance = MercuryInterface(redirection='none')
        instance.initialize_code()
        instance.set_central_particle_state(DUMMYID, 1., 2., 3., 4., 5., 6., 7., 8.)
        mass, radius, j2, j4, j6, lx, ly, lz, err = instance.get_central_particle_state(DUMMYID)
        self.assertEqual(mass, 1.0)
        self.assertEqual(radius, 2.0)
        self.assertEqual(j2, 3.0)
        self.assertEqual(j4, 4.0)
        self.assertEqual(j6, 5.0)
        self.assertEqual(lx, 6.0)
        self.assertEqual(ly, 7.0)
        self.assertEqual(lz, 8.0)
        instance.stop()

    def test15(self):
        instance = MercuryInterface(redirection='none')
        instance.initialize_code()
        i, err = instance.get_integrator()
        self.assertEqual(i, 10)
        err = instance.set_integrator(2)
        self.assertEqual(err, 0)
        i, err = instance.get_integrator()
        self.assertEqual(i, 2)
        err = instance.set_integrator(22)
        self.assertEqual(err, -1)
        instance.stop()

    def test16(self):
        instance = MercuryInterface()
        instance.initialize_code()
        err = instance.set_integrator(2)
        mass = 3.04043264264672381E-06
        dens = 5.52
        x = 2.42093942183383037E-01
        y = -9.87467766698604366E-01
        z = -4.54276292555233496E-06
        vx = 1.64294055023289365E-02
        vy = 4.03200725816140870E-03
        vz = 1.13609607260006795E-08
        sx = sy = sz = 0.
        celimit = 20.
        pid, err = instance.new_orbiter(mass, dens, x, y, z, vx, vy, vz, sx, sy, sz, celimit)
        instance.commit_particles()
        instance.enable_stopping_condition(3)
        instance.set_stopping_condition_timeout_parameter(0.01)
        etot1, err = instance.get_total_energy()
        err = instance.evolve_model(8*1e100)
        self.assertTrue(instance.is_stopping_condition_set(3)['result'] == 1)
        etot2, err = instance.get_total_energy()
        self.assertAlmostEqual(etot1, etot2, 7)
        instance.stop()


class TestMercury(TestWithMPI):

    def setUp(self):
        super(TestWithMPI, self).setUp()

    def sun_and_earth(self):
        orbiter = datamodel.Particles(1)
        orbiter.mass = 5.97e24 | units.kg
        orbiter.density = 1.0 | units.g/units.cm**3
        orbiter.position = [1.0, 0.0, 0.0] | units.AU
        orbiter.velocity = [0.0, 2.0*3.1415926535*1.0/365, 0.0] | units.AUd
        orbiter.angular_momentum = [1.0, 0, 0] | units.MSun * units.AU**2/units.day
        orbiter.celimit = 0.0 | units.none

        centre = datamodel.Particles(1)
        centre.mass = 1.0 | units.MSun
        centre.radius = .0000001 | units.AU
        centre.j2 = .0001 | units.AU**2
        centre.j4 = .0 | units.AU**4
        centre.j6 = .0 | units.AU**6

        centre.angular_momentum = [0.0, 0.0, 0.0] | units.MSun * units.AU**2/units.day

        return centre, orbiter

    def test0(self):
        centre, orbiter = self.sun_and_earth()
        mercury = MercuryWayWard()  # ,debugger='xterm')
        mercury.initialize_code()
        mercury.central_particle.add_particles(centre)
        mercury.orbiters.add_particles(orbiter)
        mercury.commit_particles()

        self.assertAlmostEqual(mercury.central_particle.j4, .0 | units.AU**4)
        self.assertAlmostEqual(mercury.central_particle.mass, 1.98892e+30 | units.kg, 3)
        self.assertAlmostEqual(mercury.central_particle.mass, 1.0 | units.MSun, 3)
        self.assertEqual(mercury.get_number_of_orbiters(), 1)
        self.assertEqual(mercury.orbiters.position, [[1, 0, 0]] | units.AU)
        self.assertEqual(mercury.orbiters.density, 1.0 | units.g/units.cm**3)
        self.assertEqual(mercury.orbiters.angular_momentum, [[1.0, 0.0, 0.0]] | units.MSun*units.AU**2/units.day)

        mercury.evolve_model(365.24 | units.day)

        self.assertAlmostEqual(mercury.orbiters.position, [[1.0, 0.0, 0.0]] | units.AU, 1)
        self.assertAlmostEqual(mercury.kinetic_energy+mercury.potential_energy, mercury.total_energy, 3)

        mercury.stop()

    def test1(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()
        mercury.initialize_code()

        mercury.central_particle.add_particles(centre)
        mercury.orbiters.add_particles(orbiters)
        mercury.commit_particles()

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(365.14 | units.day)
        self.assertAlmostEqual(mercury.orbiters[2].position, start_pos, 1)
        mercury.stop()

    def test2(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()
        mercury.initialize_code()

        mercury.central_particle.add_particles(centre)
        mercury.orbiters.add_particles(orbiters)
        mercury.commit_particles()

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(16. | units.day)
        self.assertEqual(mercury.get_time(), 16. | units.day)
        mercury.stop()

    def test3(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.initialize_code()

        mercury.particles.add_particles(solsys)
        mercury.commit_particles()

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(365.14 | units.day)
        self.assertAlmostEqual(mercury.orbiters[2].position, start_pos, 1)
        mercury.stop()

    def test4(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.initialize_code()
        self.assertEqual(mercury.parameters.timestep, 8 | units.day)
        mercury.set_initial_timestep(1 | units.day)
        mercury.parameters.timestep = 1 | units.day

        mercury.particles.add_particles(solsys)
        mercury.commit_particles()

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(365.14 | units.day)
        self.assertAlmostEqual(mercury.orbiters[2].position, start_pos, 1)
        mercury.stop()

    def test5(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()
        self.assertEqual(mercury.get_name_of_current_state(), 'UNINITIALIZED')
        mercury.initialize_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'INITIALIZED')
        mercury.commit_parameters()
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')

        mercury.central_particle.add_particles(centre)
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.orbiters.add_particles(orbiters)
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')

        mercury.commit_particles()
        self.assertEqual(mercury.get_name_of_current_state(), 'RUN')

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(365.14 | units.day)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')

        self.assertAlmostEqual(mercury.orbiters[2].position, start_pos, 1)
        mercury.cleanup_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'END')
        mercury.stop()

    def test6(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()
        self.assertEqual(mercury.get_name_of_current_state(), 'UNINITIALIZED')
        mercury.initialize_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'INITIALIZED')
        mercury.commit_parameters()
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')

        mercury.central_particle.add_particles(centre)
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.orbiters.add_particles(orbiters[4:5])
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.commit_particles()
        self.assertEqual(mercury.get_name_of_current_state(), 'RUN')

        start_pos = mercury.orbiters[0].position
        mercury.evolve_model(11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

        mercury.orbiters.add_particles(orbiters[0:1])
        self.assertEqual(mercury.get_name_of_current_state(), 'UPDATE')
        mercury.recommit_particles()

        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

        start_pos = mercury.orbiters[0].position
        mercury.evolve_model(2*11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

        mercury.cleanup_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'END')
        mercury.stop()

    def test7(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()  # debugger="gdb")
        self.assertEqual(mercury.get_name_of_current_state(), 'UNINITIALIZED')
        mercury.initialize_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'INITIALIZED')
        mercury.commit_parameters()
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')

        mercury.central_particle.add_particles(centre)
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.orbiters.add_particles(orbiters[0:5])
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.commit_particles()
        self.assertEqual(mercury.get_name_of_current_state(), 'RUN')

        start_pos = mercury.orbiters[4].position
        mercury.evolve_model(11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[4].position, start_pos, 1)

        mercury.orbiters.remove_particles(orbiters[0:1])
        self.assertEqual(mercury.get_name_of_current_state(), 'UPDATE')
        mercury.recommit_particles()

        self.assertAlmostEqual(mercury.orbiters[3].position, start_pos, 1)

        start_pos = mercury.orbiters[3].position
        mercury.evolve_model(2*11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[3].position, start_pos, 1)

        mercury.cleanup_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'END')
        mercury.stop()

    def test8(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()  # debugger="gdb")
        self.assertEqual(mercury.get_name_of_current_state(), 'UNINITIALIZED')
        mercury.initialize_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'INITIALIZED')
        mercury.commit_parameters()
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')

        mercury.central_particle.add_particles(centre)
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.orbiters.add_particles(orbiters[0:5])
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.commit_particles()
        self.assertEqual(mercury.get_name_of_current_state(), 'RUN')

        start_pos = mercury.orbiters[4].position
        mercury.evolve_model(11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[4].position, start_pos, 1)

        mercury.orbiters.remove_particles(orbiters[0:4])
        self.assertEqual(mercury.get_name_of_current_state(), 'UPDATE')
        mercury.recommit_particles()

        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

        start_pos = mercury.orbiters[0].position
        mercury.evolve_model(2*11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

        mercury.cleanup_code()
        self.assertEqual(mercury.get_name_of_current_state(), 'END')
        mercury.stop()

    def test9(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()  # debugger="gdb")
        self.assertEqual(mercury.get_name_of_current_state(), 'UNINITIALIZED')
        mercury.central_particle.add_particles(centre)
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.orbiters.add_particles(orbiters[0:5])
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        start_pos = mercury.orbiters[4].position
        mercury.evolve_model(11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[4].position, start_pos, 1)

        mercury.orbiters.remove_particles(orbiters[0:4])
        self.assertEqual(mercury.get_name_of_current_state(), 'UPDATE')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

        start_pos = mercury.orbiters[0].position
        mercury.evolve_model(2*11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

    def test10(self):
        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()
        mercury.central_particle.add_particles(centre)
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        mercury.orbiters.add_particles(orbiters[4:5])
        self.assertEqual(mercury.get_name_of_current_state(), 'EDIT')
        start_pos = mercury.orbiters[0].position
        mercury.evolve_model(11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

        mercury.orbiters.add_particles(orbiters[0:4])
        self.assertEqual(mercury.get_name_of_current_state(), 'UPDATE')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)
        start_pos = mercury.orbiters[0].position
        mercury.evolve_model(2*11.8618 | units.yr)
        self.assertEqual(mercury.get_name_of_current_state(), 'EVOLVED')
        self.assertAlmostEqual(mercury.orbiters[0].position, start_pos, 1)

    def test11(self):
        solsys = new_solar_system()

        mercury = Mercury()
        self.assertEqual(mercury.parameters.timestep, 8 | units.day)
        mercury.parameters.timestep = 1 | units.day
        self.assertEqual(mercury.parameters.timestep, 1 | units.day)

        mercury.particles.add_particles(solsys)
        start_pos = mercury.particles[5].position
        mercury.evolve_model(11.8618 | units.yr)
        self.assertAlmostEqual(mercury.particles[5].position, start_pos, 1)
        mercury.stop()

    def test12(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.parameters.timestep = 1. | units.day

        mercury.particles.add_particles(solsys)
        start_pos = mercury.particles[5].position
        mercury.evolve_model(11.8618 | units.yr)
        self.assertAlmostEqual(mercury.particles[5].position, start_pos, 1)
        mercury.particles.remove_particles(mercury.particles[1:5])
        self.assertAlmostEqual(mercury.particles[1].position, start_pos, 1)
        start_pos = mercury.particles[1].position
        mercury.evolve_model(2*11.8618 | units.yr)
        self.assertAlmostEqual(mercury.particles[1].position, start_pos, 1)
        mercury.stop()

    def test13(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.particles.add_particles(solsys)
        idpos1 = [(p.position - q.position) for p in mercury.particles[1:10] for q in mercury.particles[1:10]]
        mercury.evolve_model(11.8618 | units.yr)
        edpos1 = [(p.position - q.position) for p in mercury.particles[1:10] for q in mercury.particles[1:10]]
        mercury.stop()

        centre, orbiters = new_solar_system_for_mercury()

        mercury = MercuryWayWard()
        mercury.central_particle.add_particles(centre)
        mercury.orbiters.add_particles(orbiters)

        idpos2 = [(p.position - q.position) for p in mercury.orbiters[0:9] for q in mercury.orbiters[0:9]]
        mercury.evolve_model(11.8618 | units.yr)
        edpos2 = [(p.position - q.position) for p in mercury.orbiters[0:9] for q in mercury.orbiters[0:9]]
        mercury.stop()

        for d1, d2 in zip(idpos1, idpos2):
            self.assertAlmostEqual(d1, d2, 7)
        for d1, d2 in zip(edpos1, edpos2):
            self.assertAlmostEqual(d1, d2, 7)

    def test14(self):
        centre, orbiters = new_solar_system_for_mercury()
        oneyear = 365.14 | units.day
        halfyear = oneyear / 2.0
        mercury = MercuryWayWard()
        mercury.initialize_code()
        mercury.commit_parameters()

        mercury.central_particle.add_particles(centre)
        mercury.orbiters.add_particles(orbiters)

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(halfyear)
        central_particles = mercury.central_particle.copy()
        orbiters = mercury.orbiters.copy()
        mercury.stop()

        mercury = MercuryWayWard()
        mercury.initialize_code()
        mercury.parameters.begin_time = halfyear
        mercury.commit_parameters()

        mercury.central_particle.add_particles(centre)
        mercury.orbiters.add_particles(orbiters)
        mercury.evolve_model(oneyear)
        self.assertAlmostEqual(mercury.orbiters[2].position, start_pos, 1)
        mercury.stop()

    def test15(self):
        solsys = new_solar_system()

        solsys.x -= 1. | units.AU

        p = datamodel.Particles(3)
        p.mass = [1, 2, 3] | units.MSun
        p.x = [1, 10, 100] | units.AU
        p.y = [0, 0, -10] | units.AU
        p.z = [0, 0, 10] | units.AU

        pe = p.potential_energy_in_field(solsys)

        mercury = Mercury()
        mercury.particles.add_particles(solsys)
        pot = mercury.get_potential_at_point(p.x*0., p.x, p.y, p.z)

        self.assertAlmostRelativeEqual((pot*p.mass).sum(), pe, 12)

    def test16(self):
        solsys = new_solar_system()

        solsys.x -= 1. | units.AU

        p = datamodel.Particles(3)
        p.mass = [1, 2, 3] | units.MSun
        p.x = [1, 10, 100] | units.AU
        p.y = [0, 0, -10] | units.AU
        p.z = [0, 0, 10] | units.AU

        from amuse.community.huayno.interface import Huayno
        from amuse.units import nbody_system
        conv = nbody_system.nbody_to_si(1. | units.AU, 1. | units.MSun)
        h = Huayno(conv)
        h.particles.add_particles(solsys)

        ax1, ay1, az1 = h.get_gravity_at_point(p.x*0., p.x, p.y, p.z)

        mercury = Mercury()
        mercury.particles.add_particles(solsys)

        ax2, ay2, az2 = mercury.get_gravity_at_point(p.x*0., p.x, p.y, p.z)

        self.assertAlmostRelativeEqual(ax1, ax2, 12)
        self.assertAlmostRelativeEqual(ay1, ay2, 12)
        self.assertAlmostRelativeEqual(az1, az2, 12)

    def test17(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.particles.add_particles(solsys)
        mercury.evolve_model(5. | units.yr)
        mercury.evolve_model(10. | units.yr)

        dpos1 = mercury.particles[0].position-mercury.particles[5].position

        mercury.stop()

        mercury = Mercury()
        mercury.particles.add_particles(solsys)
        mercury.evolve_model(5. | units.yr)
        mercury.particles.x += 1 | units.AU
        mercury.particles.y += 2 | units.AU
        mercury.particles.z += 3 | units.AU
        mercury.particles.vx += 10. | units.kms
        mercury.particles.vy += 20. | units.kms
        mercury.particles.vz += 30. | units.kms
        mercury.evolve_model(10. | units.yr)

        dpos2 = mercury.particles[0].position-mercury.particles[5].position
        mercury.stop()

        self.assertAlmostEqual(dpos1, dpos2, 12)

    def test18(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.initialize_code()
        self.assertEqual(mercury.parameters.integrator, 10)
        mercury.parameters.integrator = 2
        self.assertEqual(mercury.parameters.integrator, 2)
        mercury.stop()

    def test19(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.initialize_code()
        mercury.parameters.integrator = 1

        mercury.particles.add_particles(solsys)
        mercury.commit_particles()

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(365.14 | units.day)
        self.assertAlmostEqual(mercury.orbiters[2].position, start_pos, 1)
        mercury.stop()

    def test20(self):
        solsys = new_solar_system()

        mercury = Mercury()  # debugger="gdb")
        mercury.initialize_code()
        mercury.parameters.integrator = 2

        mercury.particles.add_particles(solsys)
        mercury.commit_particles()

        start_pos = mercury.orbiters[2].position
        mercury.evolve_model(365.14 | units.day)
        self.assertAlmostEqual(mercury.orbiters[2].position, start_pos, 1)
        mercury.stop()

    def test21(self):
        solsys = new_solar_system()

        mercury = Mercury()
        mercury.initialize_code()

        names = ["elements_file", "close_encounters_file", "info_file",
         "bigbody_file", "smallbody_file", "integration_parameters_file", "restart_file"]

        for name in names:
            self.assertEqual(getattr(mercury.parameters, name), "/dev/null")

        for name in names:
            setattr(mercury.parameters, name, os.path.join(mercury.output_directory, name))

        for name in names:
            self.assertEqual(getattr(mercury.parameters, name), os.path.join(mercury.output_directory, name))

        mercury.stop()

    def xtest22(self):
        """ collision test hangs or fails if internal collision detection is enabled """

        def collision():

            M = 1. | units.MSun
            m = 1. | units.MJupiter
            r = 5. | units.AU
            vcirc = (constants.G*(M+m)/r)**0.5

            sys = datamodel.Particles(4)
            sys[0].mass = M
            sys[0].radius = 1. | units.RSun
            sys[0].x = 0 | units.AU
            sys[0].y = 0 | units.AU
            sys[0].z = 0 | units.AU
            sys[0].vx = 0 | units.kms
            sys[0].vy = 0 | units.kms
            sys[0].vz = 0 | units.kms

            sys[1].mass = m
            sys[1].radius = 0.01 | units.RSun
            sys[1].x = r
            sys[1].y = 0 | units.AU
            sys[1].z = 0 | units.AU
            sys[1].vx = 0 | units.kms
            sys[1].vy = vcirc
            sys[1].vz = 0 | units.kms

            sys[2].mass = m
            sys[2].radius = 0.01 | units.RSun
            sys[2].x = -r
            sys[2].y = 0 | units.AU
            sys[2].z = 0 | units.AU
            sys[2].vx = 0 | units.kms
            sys[2].vy = vcirc
            sys[2].vz = 0 | units.kms

            sys[3].mass = m
            sys[3].radius = 0.01 | units.RSun
            sys[3].x = 0 | units.AU
            sys[3].y = r
            sys[3].z = 0 | units.AU
            sys[3].vx = 0 | units.kms
            sys[3].vy = 0 | units.kms
            sys[3].vz = 0 | units.kms

            return sys

        code = Mercury()

        sys = collision()

        code.particles.add_particles(sys)

        tend = 3.5 | units.yr
        dt = 100. | units.day
        tnow = code.model_time

        while tnow < tend:
            code.evolve_model(tnow+dt)
            tnow = code.model_time
            print(tnow.in_(units.yr))

        code.stop()



import os.path
import io
import numpy

from amuse import io
from amuse.io import starlab
from amuse.units import units
from amuse.units import nbody_system
from amuse.test import amusetest
from amuse import datamodel
plummer_scaled_content = """(Particle
  N = 5
(Log
 ===>  Wed Mar 23 10:36:19 2011
       Starlab 4.4.4 (user vanelteren) : makeplummer -n 5
       random number generator seed = 1300872979
  initial_mass = 1.93791189968789
 ===>  Wed Mar 23 10:36:19 2011
       Starlab 4.4.4 (user vanelteren) : makemass -l 0.2 -u 20 -x -2.35
       random number generator seed = 1300872979
       Power_Law mass function, total mass =     1.94 Solar
)Log
(Dynamics
  system_time  =  0
  m  =  1.93791189968789257
  r  =  0 0 0
  v  =  0 0 0
  com_time = 0
  com_pos = 0 0 0
  com_vel = 0 0 0
  total_energy = -0.25
)Dynamics
(Hydro
)Hydro
(Star
  mass_scale     =  0.516019329960796136
  size_scale     =  -2.25500000000000001e-08
  time_scale     =  -1
)Star
(Particle
  N = 1
(Log
)Log
(Dynamics
  m  =  0.266730587350341442
  r  =  -0.934040458729136547 -0.695803261872860679 0.564081767628105579
  v  =  0.293044501505944688 0.0259404966497079996 0.196831834670057881
)Dynamics
(Hydro
)Hydro
(Star
)Star
)Particle
(Particle
  N = 1
(Log
)Log
(Dynamics
  m  =  0.32415747586178062
  r  =  -0.177487175943360831 0.205223807753114523 -0.191956558941283828
  v  =  -0.178395285089674116 -0.0357730795197053753 0.376231470164175796
)Dynamics
(Hydro
)Hydro
(Star
)Star
)Particle
(Particle
  N = 1
(Log
)Log
(Dynamics
  m  =  0.567097501003086424
  r  =  0.607757256863783235 0.120278701131815768 0.338645014325028748
  v  =  -0.26638454687085783 0.291986511517820568 0.672417896548585303
)Dynamics
(Hydro
)Hydro
(Star
)Star
)Particle
(Particle
  N = 1
(Log
)Log
(Dynamics
  m  =  0.379463033750591316
  r  =  0.247202300114546969 0.086741731241469916 -0.164892802949227257
  v  =  -0.0888917818811222754 0.286288450446576914 -1.03818564484543829
)Dynamics
(Hydro
)Hydro
(Star
)Star
)Particle
(Particle
  N = 1
(Log
)Log
(Dynamics
  m  =  0.400463301722092879
  r  =  0.256568077694167229 0.283559021746460527 -0.545877420062623298
  v  =  0.240627112335709492 -0.568442379094400096 -0.207295556537380576
)Dynamics
(Hydro
)Hydro
(Star
)Star
)Particle
)Particle
"""

with_stellar_structure = """(Particle
  N = 2
(Log
 ===>  Wed Mar 23 12:28:21 2011
       Starlab 4.4.4 (user vanelteren) : makeplummer -n 2
       random number generator seed = 1300879701
  initial_mass = 1
 ===>  Wed Mar 23 12:28:21 2011
       Starlab 4.4.4 (user vanelteren) : makemass -f 1 -x -2.0 -l 0.1 -u 20
       random number generator seed = 1300879701
       Power_Law mass function, total mass =     0.22 Solar
 ===>  Wed Mar 23 12:28:21 2011
       Starlab 4.4.4 (user vanelteren) : add_star -Q 0.5 -R 5
 ===>  Wed Mar 23 12:28:21 2011
       Starlab 4.4.4 (user vanelteren) : scale -s
  initial_total_energy = -0.25
  initial_rvirial = 1
)Log
(Dynamics
  system_time  =  0
  m  =  1.000000000000002
  r  =  0 0 0
  v  =  0 0 0
  com_time = 0
  com_pos = 0 0 0
  com_vel = 0 0 0
  total_energy = -0.25
)Dynamics
(Hydro
)Hydro
(Star
  mass_scale     =  4.62223018200823255
  size_scale     =  4.51000000000000035e-09
  time_scale     =  0.0028016566610805007
)Star
(Particle
  N = 1
(Log
)Log
(Dynamics
  m  =  0.483839787917132669
  r  =  -0.209118735762131774 -0.0880146969484976449 0.122429686361064466
  v  =  0.435701422371714053 -0.0578440884891995299 0.583282312300410277
)Dynamics
(Hydro
)Hydro
(Star
  Type   =  main_sequence
  T_cur  =  0
  M_rel  =  0.104676696933106134
  M_env  =  0.0946766969331061387
  M_core =  0.0100000000000000002
  T_eff  =  3011.01000587455155
  L_eff  =  0.00122847524117014736
)Star
)Particle
(Particle
  N = 1
(Log
)Log
(Dynamics
  m  =  0.516160212082869219
  r  =  0.196024339714903045 0.0825034772310478115 -0.114763501907016688
  v  =  -0.408419089384746692 0.0542220629403740648 -0.546758900962974193
)Dynamics
(Hydro
)Hydro
(Star
  Type   =  main_sequence
  T_cur  =  0
  M_rel  =  0.111669084350665276
  M_env  =  0.101669084350665281
  M_core =  0.0100000000000000002
  T_eff  =  3051.91244441666595
  L_eff  =  0.00147739782384390963
)Star
)Particle
)Particle
"""


class Test(amusetest.TestCase):

    def test1(self):
        """test_starlab.test1

                +---------------------------------------------------------------------+   
                |                      Particle tree of test_subub.dyn                |   
                |                                                                     |   
                |                           0 16kg x=0                                |   
                |                           ,-  .                                     |   
                |                        ,-'     `._                                  |   
                |                     _,'           `.                                |   
                |                   ,'                `-.                             |   
                |                 1 8kg, x=-10        3 8kg, x=10                     |   
                |                ,-'                   - ._                           |   
                |              ,'                    ,'    `.                         |   
                |            ,'                    ,'        `._                      |   
                |          ,'                    ,'             `.                    |   
                |       2  4kg, x=-15          4 4kg, x=5      5 4kg, x=15            |   
                |                                                   .                 |   
                |                                                    `._              |   
                |                                                       `.            |   
                |                                                         `           |   
                |                                                       6 2kg, x=17   |   
                |                                                                     |   
                +---------------------------------------------------------------------+   

        """
        directory = os.path.dirname(__file__)
        convert_nbody = nbody_system.nbody_to_si(1 | units.kg, 1 | units.m)

        I = starlab.ParticlesFromDyn(os.path.join(directory, 'test_subsub.dyn'))

        All = I.Particles

        self.assertEqual(len(All), 7)
        self.assertEqual(len(All[0].descendents()), 6)
        self.assertEqual(All[0].children().mass.value_in(nbody_system.mass)[0], 8.0)
        self.assertEqual(All[1].children().mass.value_in(nbody_system.mass)[0], 4.0)
        self.assertEqual(All[5].children().mass.value_in(nbody_system.mass)[0], 2.0)

    def test2(self):
        directory = os.path.dirname(__file__)
        convert_nbody = nbody_system.nbody_to_si(1 | units.kg, 1 | units.m)

        I = starlab.ParticlesFromDyn(os.path.join(directory, 'test_subsub.dyn'), convert_nbody)

        All = I.Particles

        self.assertEqual(len(All), 7)
        self.assertEqual(len(All[0].descendents()), 6)
        self.assertEqual(All[0].children().mass.value_in(units.kg)[0], 8.0)
        self.assertEqual(All[1].children().mass.value_in(units.kg)[0], 4.0)
        self.assertEqual(All[5].children().mass.value_in(units.kg)[0], 2.0)

    def test3(self):
        directory = os.path.dirname(__file__)
        set = io.read_set_from_file(os.path.join(directory, 'plummer.dyn'), 'starlab')
        self.assertEqual(len(set), 10)
        self.assertAlmostRelativeEquals(set.mass, 0.1 | nbody_system.mass)

    def test4(self):
        particle_set = starlab.StarlabFileFormatProcessor().load_string(plummer_scaled_content)
        self.assertEqual(len(particle_set), 5)
        print(particle_set.mass.as_quantity_in(units.MSun))
        print(particle_set.mass.as_quantity_in(units.kg))
        print(particle_set)
        self.assertTrue(numpy.all(particle_set.mass > 0.2 | units.MSun))
        self.assertTrue(numpy.all(particle_set.mass < 1.1 | units.MSun))
        self.assertTrue(numpy.all(particle_set.x > -1 | units.parsec))
        self.assertTrue(numpy.all(particle_set.x < 1 | units.parsec))
        self.assertTrue(numpy.all(particle_set.y > -1 | units.parsec))
        self.assertTrue(numpy.all(particle_set.y < 1 | units.parsec))
        self.assertTrue(numpy.all(particle_set.z > -1 | units.parsec))
        self.assertTrue(numpy.all(particle_set.z < 1 | units.parsec))

    def test5(self):
        set = starlab.StarlabFileFormatProcessor().load_string(with_stellar_structure)
        self.assertEqual(len(set), 2)
        self.assertAlmostRelativeEquals(set[0].envelope_mass, 0.0946766969331061387 | units.MSun)
        self.assertAlmostRelativeEquals(set[0].core_mass, 0.0100000000000000002 | units.MSun)
        self.assertAlmostRelativeEquals(set[0].relative_mass,  set[0].mass, 10)
        self.assertAlmostRelativeEquals(set[0].effective_temperature,  3011.01000587455155 | units.K)
        self.assertAlmostRelativeEquals(set[0].effective_luminocity,  0.00122847524117014736 | units.LSun)
        self.assertAlmostRelativeEquals(set[0].stellar_type, units.stellar_type("Main Sequence star"))
        self.assertAlmostRelativeEquals(set[1].relative_mass,  set[1].mass, 10)

    def test6(self):
        directory = os.path.dirname(__file__)
        set = io.read_set_from_file(os.path.join(directory, 'evolved.dyn'), 'starlab')
        self.assertEqual(len(set), 20)

        self.assertAlmostRelativeEquals(set.time, set.age, 4)  # only accurate to 4, starlab stellar evolution time scale?

        self.assertAlmostRelativeEquals(set[0].velocity, [177.579717905, 38.5027308364, -35.8571344243] | units.km / units.hour, 8)
        self.assertAlmostRelativeEquals(set[0].acceleration, [-0.000648471729782, 0.000309476774701, -0.000356623346185] | units.parsec / (units.Myr ** 2), 8)
        self.assertAlmostRelativeEquals(set.unconverted_set()[0].specific_potential, -0.32735384622167929 | nbody_system.potential)
        # select the main sequence star, the dwarf masses don't match
        main_sequence_stars = set.select(lambda stellar_type: stellar_type == units.stellar_type("Main Sequence star"), ["stellar_type"])
        self.assertAlmostRelativeEquals(main_sequence_stars.mass, main_sequence_stars.relative_mass, 10)
        carbon_dwarfs = set.select(lambda stellar_type: stellar_type == units.stellar_type("Carbon/Oxygen White Dwarf"), ["stellar_type"])
        self.assertAlmostRelativeEquals(carbon_dwarfs.mass, carbon_dwarfs.core_mass, 10)

    def test7(self):
        directory = os.path.dirname(__file__)
        set, converter = io.read_set_from_file(os.path.join(directory, 'evolved.dyn'), 'starlab', return_converter=True)
        self.assertEqual(len(set), 20)

        unconverted_set = io.read_set_from_file(os.path.join(directory, 'evolved.dyn'), 'starlab', must_scale=False)
        self.assertEqual(len(unconverted_set), 20)
        self.assertAlmostRelativeEquals(converter.to_nbody(set.x), unconverted_set.x)



import numpy
from amuse.test import amusetest
from amuse.units import units, quantities

from amuse.ext import static_potentials

from amuse.support.console import set_printing_strategy


class TestGalacticCenterPotential(amusetest.TestCase):
    def setUp(self):
        set_printing_strategy("custom", preferred_units=[units.MSun, units.parsec, units.Myr, units.kms, units.kms**2, units.m/units.s**2])

    def tearDown(self):
        set_printing_strategy("default")

    def test1(self):
        """ test enclosed mass profile of Kruijssen potential """

        static_potential = static_potentials.Galactic_Center_Potential_Kruijssen()

        m_zero = static_potential.enclosed_mass(0 | units.parsec)
        self.assertAlmostEqual(m_zero, 0 | units.MSun)

        factor = 8.3/8.5
        m_on_bin = static_potential.enclosed_mass(7.489 | units.parsec * factor)
        self.assertAlmostEqual(m_on_bin, factor**2 * 17170000. | units.MSun)

        m_in_bin = static_potential.enclosed_mass(200 | units.parsec)
        self.assertAlmostEqual(m_in_bin, 1.644965507 | 1E9*units.MSun)

        # test array
        r = numpy.logspace(1, 3, 500) | units.parsec
        m = static_potential.enclosed_mass(r)
        self.assertAlmostEqual(m[10], 2.0680055041 | 1E7*units.MSun)

        # test 2D array
        x = numpy.logspace(1, 3, 500) | units.parsec
        matrix, m = quantities.meshgrid(x, x)
        mm = static_potential.enclosed_mass(matrix)
        self.assertAlmostEqual(mm[20, 20], 2.223803269 | 1E7*units.MSun)

        # quick plot for testing
        # from matplotlib import pyplot
        # from amuse import plot as aplot
        # aplot.loglog(r, m, marker="*")
        # aplot.scatter(static_potential.radius, static_potential.enclosed_mass_profile)
        # pyplot.show()

    def test2(self):
        """ Test get_potential_at_point """
        static_potential = static_potentials.Galactic_Center_Potential_Kruijssen()

        points = [[1, 0, 0],
                  [0, 10, 0],
                  [1e3, 0, 1e2],
                  ] | units.parsec
        potentials = static_potential.get_potential_at_point(0, *points.transpose())
        # print potentials.in_(units.kms**2)
        self.assertAlmostEqual(potentials[0], -17280.958078 | units.kms**2, 6)
        self.assertAlmostEqual(potentials[1], -8296.5300805 | units.kms**2, 6)
        self.assertAlmostEqual(potentials[2], -14965.2205261 | units.kms**2, 6)

    def test3(self):
        """ Test get_gravity_at_point """
        static_potential = static_potentials.Galactic_Center_Potential_Kruijssen()

        points = [[1, 0, 0],
                  [0, 10, 0],
                  [1e3, 0, 1e2],
                  ] | units.parsec
        grav = static_potential.get_gravity_at_point(0, *points.transpose()).transpose()
        print(grav)
        self.assertAlmostEqual(grav[0], [-5.6003771, 0, 0] | 1e-7*units.m/units.s**2, 6)
        self.assertAlmostEqual(grav[1], [0, -2.6887222, 0] | 1e-8*units.m/units.s**2, 6)
        self.assertAlmostEqual(grav[2], [-4.7661597, 0, -1.200846] | 1e-10*units.m/units.s**2, 6)


class TestGalacticPotential(amusetest.TestCase):
    def test1(self):
        """ Test the Galactic potential """
        static_potential = static_potentials.Disc_Bulge_Halo_Potential()

        points = [[1, 0, 0],
                  [10, 0, 0],
                  [1e3, 0, 1e4],
                  ] | units.kpc
        potentials = static_potential.get_potential_at_point(0, *points.transpose()).transpose()

        self.assertAlmostEqual(potentials[0], -96648.29247 | units.kms**2, 6)
        self.assertAlmostEqual(potentials[1], 8827.40917774 | units.kms**2, 6)
        self.assertAlmostEqual(potentials[2], 266129.25389 | units.kms**2, 6)



from amuse.test import amusetest

from amuse.support.exceptions import AmuseException
from amuse.units import units
from amuse.units import nbody_system
from amuse.units import constants
from amuse.support import console
from amuse.support.console import set_printing_strategy
from amuse.support.console import get_current_printing_strategy

import sys


class TestPrintingStrategy(amusetest.TestCase):

    def tearDown(self):
        set_printing_strategy("default")

    def test1(self):
        print("Testing get/set of printing strategy")
        self.assertEqual(get_current_printing_strategy(), console.DefaultPrintingStrategy)
        set_printing_strategy("no_unit")
        self.assertEqual(get_current_printing_strategy(), console.NoUnitsPrintingStrategy)
        set_printing_strategy("default")
        self.assertEqual(get_current_printing_strategy(), console.DefaultPrintingStrategy)
        set_printing_strategy("no_units")
        self.assertEqual(get_current_printing_strategy(), console.NoUnitsPrintingStrategy)
        set_printing_strategy("with_units")
        self.assertEqual(get_current_printing_strategy(), console.DefaultPrintingStrategy)
        set_printing_strategy("formal")
        self.assertEqual(get_current_printing_strategy(), console.FormalPrintingStrategy)
        set_printing_strategy("nbody")
        self.assertEqual(get_current_printing_strategy(), console.NBodyPrintingStrategy)
        set_printing_strategy(console.NoUnitsPrintingStrategy)
        self.assertEqual(get_current_printing_strategy(), console.NoUnitsPrintingStrategy)
        set_printing_strategy(console.DefaultPrintingStrategy)
        self.assertEqual(get_current_printing_strategy(), console.DefaultPrintingStrategy)

    def test2(self):
        print("Testing no units printing strategy with SI quantities")
        mass = 1.0 | units.kg
        acc = 9.8 | units.m / units.s**2
        position = [1, 2.0, 3] | units.m
        mass_in_g = mass.as_quantity_in(units.g)
        pi = 3.14 | units.none

        self.assertEqual(str(mass),      "1.0 kg")
        self.assertEqual(str(acc),       "9.8 m / (s**2)")
        self.assertEqual(str(position),  "[1.0, 2.0, 3.0] m")
        self.assertEqual(str(mass_in_g), "1000.0 g")
        self.assertEqual(str(pi),        "3.14 none")

        set_printing_strategy("no_unit")
        self.assertEqual(get_current_printing_strategy(), console.NoUnitsPrintingStrategy)
        self.assertEqual(str(mass),      "1.0")
        self.assertEqual(str(acc),       "9.8")
        self.assertEqual(str(position),  "[1.0, 2.0, 3.0]")
        self.assertEqual(str(mass_in_g), "1000.0")
        self.assertEqual(str(pi),        "3.14")
        set_printing_strategy("default")

    def test3(self):
        print("Testing no units printing strategy with N-body quantities")
        mass = 1.0 | nbody_system.mass
        acc = 9.8 | nbody_system.acceleration
        position = [1, 2, 3] | nbody_system.length

        self.assertEqual(str(mass),      "1.0 mass")
        self.assertEqual(str(acc),       "9.8 length / (time**2)")
        self.assertEqual(str(position),  "[1, 2, 3] length")

        set_printing_strategy("no_unit")
        self.assertEqual(str(mass),      "1.0")
        self.assertEqual(str(acc),       "9.8")
        self.assertEqual(str(position),  "[1, 2, 3]")
        set_printing_strategy("default")

    def test4(self):
        print("Testing formal printing strategy")
        mass = 1.0 | units.kg
        acc = 9.8 | units.m / units.s**2
        position = [1, 2.0, 3] | nbody_system.length
        mass_in_g = mass.as_quantity_in(units.g) * 1.0
        pi = 3.14 | units.none

        set_printing_strategy("formal")
        self.assertEqual(str(mass),      "<quantity 1.0 | kg>")
        self.assertEqual(str(acc),       "<quantity 9.8 | m / (s**2)>")
        self.assertEqual(str(position),  "<quantity [1.0, 2.0, 3.0] | length>")
        self.assertEqual(str(mass_in_g), "<quantity 1000.0 | 0.001 * kg>")  # <quantity 1000.0 | g>")
        self.assertEqual(str(pi),        "<quantity 3.14 | none>")
        set_printing_strategy("default")

    def test5(self):
        print("Testing nbody printing strategy")
        mass = 1.0 | nbody_system.mass
        acc = 9.8 | nbody_system.length / units.s**2
        position = [1, 2, 3] | units.m
        pi = 3.14 | units.none
        converter = nbody_system.nbody_to_si(1.0 | units.kg, 1.0 | units.m)

        set_printing_strategy("nbody")
        self.assertEqual(str(mass), "1.0")
        self.assertRaises(AmuseException, str, acc, expected_message="Unable to convert length / (s**2) to N-body units. No nbody_converter given")
        self.assertEqual(str(converter.to_nbody(acc * constants.G.number)), "9.8")
        if sys.hexversion > 0x03000000:
            self.assertEqual(str(converter.to_nbody(position)), "[1.0, 2.0, 3.0]")
        else:
            self.assertEqual(str(converter.to_nbody(position)), "[1, 2, 3]")
        self.assertEqual(str(pi), "3.14")

        set_printing_strategy("nbody", nbody_converter=converter)
        self.assertEqual(str(mass), "1.0")
        self.assertEqual(str(acc * constants.G.number), "9.8")
        if sys.hexversion > 0x03000000:
            self.assertEqual(str(position), "[1.0, 2.0, 3.0]")
        else:
            self.assertEqual(str(position), "[1, 2, 3]")
        set_printing_strategy("default")

    def test6(self):
        print("Testing astro printing strategy without units printed")
        mass = 2.0 | 0.5 * units.MSun
        acc = (0.0098 | nbody_system.length) * (1 | units.Myr**-2).as_quantity_in(units.s**-2)
        position = [0.1, 0.2, 0.3] | nbody_system.length
        energy = 1e8 | units.erg
        temperature = 5000 | units.K
        pi = 3.14 | units.none
        converter = nbody_system.nbody_to_si(1.0 | units.kg, 1.0 | units.kpc)

        set_printing_strategy("astro", print_units=False)
        self.assertEqual(str(mass), "1.0")
        self.assertRaises(AmuseException, str, acc, expected_message="Unable to convert length * s**-2 to SI units. No nbody_converter given")
        self.assertEqual(str(converter.to_si(acc)), "9.8")
        self.assertEqual(str(converter.to_si(position)), "[100.0, 200.0, 300.0]")
        self.assertEqual(str(energy), "10.0")
        self.assertEqual(str(constants.G)[:8], "0.00449945056135"[:8])
        self.assertEqual(str(temperature), "5000")
        self.assertEqual(str(pi), "3.14")

        set_printing_strategy("astro", nbody_converter=converter, print_units=False)
        self.assertEqual(str(acc), "9.8")
        set_printing_strategy("default")

    def test7(self):
        print("Testing astro printing strategy with units printed")
        mass = 2.0 | 0.5 * units.MSun
        acc = (0.0097 | nbody_system.length) * (1 | units.Myr**-2).as_quantity_in(units.s**-2)
        position = [0.1, 0.2, 0.3] | nbody_system.length
        energy = 1e8 | units.erg
        temperature = 5000 | units.K
        pi = 3.14 | units.none
        converter = nbody_system.nbody_to_si(1.0 | units.kg, 1.0 | units.kpc)

        set_printing_strategy("astro")
        self.assertEqual(str(mass), "1.0 MSun")
        self.assertEqual(str(acc), "0.0097 length * Myr**-2")
        self.assertEqual(str(converter.to_si(acc)), "9.7 parsec * Myr**-2")
        self.assertEqual(str(converter.to_si(position)), "[100.0, 200.0, 300.0] parsec")
        self.assertEqual(str(energy), "10.0 J")

        self.assertEqual(str(constants.G)[:8], "0.004499450561351174 parsec**3 * MSun**-1 * Myr**-2"[:8])
        self.assertEqual(str(constants.G)[-30:], "parsec**3 * MSun**-1 * Myr**-2")
        self.assertEqual(str(temperature), "5000 K")
        self.assertEqual(str(pi), "3.14 none")

        set_printing_strategy("astro", nbody_converter=converter)
        self.assertEqual(str(acc), "9.7 parsec * Myr**-2")
        set_printing_strategy("astro", ignore_converter_exceptions=False)
        self.assertRaises(AmuseException, str, acc, expected_message="Unable to convert length * s**-2 to SI units. No nbody_converter given")
        set_printing_strategy("default")

    def test8(self):
        print("Testing SI printing strategy")
        mass = 2.0 | 0.5 * units.MSun
        acc = 0.0098 | nbody_system.length / units.Myr**2
        position = [0.1, 0.2, 0.3] | nbody_system.length
        energy = 1e8 | units.erg
        temperature = 5000 | units.milli(units.K)
        pi = 3.14 | units.none
        converter = nbody_system.nbody_to_si(1.0 | units.kg, 1.0 | units.kpc)

        set_printing_strategy("SI", nbody_converter=converter)
        self.assertEqual(str(mass), "1.98892e+30 kg")
        self.assertEqual(str(acc), "3.03659755643e-10 m * s**-2")
        self.assertEqual(str(position), "[3.08567758128e+18, 6.17135516256e+18, 9.25703274384e+18] m")
        self.assertEqual(str(energy), "10.0 kg * m**2 * s**-2")
        self.assertEqual(str(constants.G), "6.67428e-11 m**3 * kg**-1 * s**-2")
        self.assertEqual(str(temperature), "5.0 K")
        self.assertEqual(str(pi), "3.14 none")

        set_printing_strategy("SI", nbody_converter=converter, print_units=False)
        self.assertEqual(str(mass), "1.98892e+30")
        self.assertEqual(str(acc), "3.03659755643e-10")
        self.assertEqual(str(position), "[3.08567758128e+18, 6.17135516256e+18, 9.25703274384e+18]")
        self.assertEqual(str(energy), "10.0")
        self.assertEqual(str(constants.G), "6.67428e-11")
        self.assertEqual(str(temperature), "5.0")
        self.assertEqual(str(pi), "3.14")
        set_printing_strategy("default")

    def test9(self):
        print("Testing custom printing strategy")
        mass = 2.0 | 0.5 * units.MSun
        acc = (0.0098 | nbody_system.length) * (1 | units.Myr**-2).as_quantity_in(units.s**-2)
        position = [0.1, 0.2, 0.3] | nbody_system.length
        power = 10 | units.W
        temperature = 5000 | units.K
        pi = 3.14 | units.none
        converter = nbody_system.nbody_to_si(1.0 | units.kg, 1.0 | units.kpc)

        set_printing_strategy("custom", nbody_converter=converter, preferred_units=[units.amu, units.au, units.minute, units.milli(units.K), units.erg], precision=3,
            prefix="(> ", separator=" <|> ", suffix=" <)")
        self.assertEqual(str(mass), "(> 1.20e+57 <|> amu <)")
        self.assertEqual(str(acc), "(> 7.31e-18 <|> au * min**-2 <)")
        self.assertEqual(str(position), "(> [2.06e+07, 4.13e+07, 6.19e+07] <|> au <)")
        self.assertEqual(str(power), "(> 6.00e+09 <|> erg / min <)")
        self.assertEqual(str(constants.G), "(> 1.19e-67 <|> au**3 * amu**-1 * min**-2 <)")
        self.assertEqual(str(constants.kB), "(> 1.38e-19 <|> erg * mK**-1 <)")
        self.assertEqual(str(temperature), "(> 5.00e+06 <|> mK <)")
        self.assertEqual(str(pi), "(> 3.14 <|> none <)")
        set_printing_strategy("default")

    def test10(self):
        print("Testing custom printing strategy with precision keyword")
        mass = 2.0 | 0.5 * units.MSun
        acc = 0.23456 | 0.54321 * units.m * units.s**-2
        velocity = [-0.12345]*3 | units.km / units.s
        position = [0.1234567890123456789, 0.2, 3.0] | units.au
        positions = [position.number]*2 | position.unit
        multi_dimensional = [positions.number]*2 | positions.unit
        pi = 3.1415926535 | units.none

        set_printing_strategy("custom", precision=3)
        self.assertEqual(str(mass), "2.00 0.5 * MSun")
        self.assertEqual(str(acc), "0.235 0.54321 * m * s**-2")
        self.assertEqual(str(velocity), "[-0.123, -0.123, -0.123] km / s")
        tmp = "[0.123, 0.200, 3.00]"
        self.assertEqual(str(position), tmp + " au")
        self.assertEqual(str(positions), "["+tmp+", "+tmp+"] au")
        self.assertEqual(str(multi_dimensional), "[["+tmp+", "+tmp +
            "], ["+tmp+", "+tmp+"]] au")
        self.assertEqual(str(pi), "3.14 none")
        set_printing_strategy("default")



from amuse.support.interface import InCodeComponentImplementation

from amuse.test.amusetest import TestWithMPI
from amuse.support import exceptions
from amuse.support import options

import os
import time
from amuse.units import nbody_system
from amuse.units import units
from amuse import datamodel
from amuse.rfi.tools import create_c
from amuse.rfi import async_request
from amuse.rfi.core import *

from . import test_c_implementation

from amuse.test import compile_tools

from amuse.community.distributed.interface import DistributedAmuse, Pilot

codestring = test_c_implementation.codestring+"""
#include <unistd.h>

float _x[10] = { 1., 2., 3., 4., 5., 6., 7., 8., 9., 10.};

int do_sleep(int in) {
    sleep(in);
    return 0;
}

int return_error(int * out) {
    *out=123;
    return -1;
}

int get_x(int in, float *x){
    *x=_x[in];
    return 0;
    }

int set_x(int in, float x){
    _x[in]=x;
    return 0;
    }

int dummy(){
    return 0;
    }

"""


class ForTestingInterface(test_c_implementation.ForTestingInterface):
    @legacy_function
    def do_sleep():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in', dtype='int32', direction=function.IN)
        function.result_type = 'int32'
        return function

    @legacy_function
    def return_error():
        function = LegacyFunctionSpecification()
        function.addParameter('out', dtype='int32', direction=function.OUT)
        function.result_type = 'int32'
        return function

    @legacy_function
    def echo_2_int():
        function = LegacyFunctionSpecification()
        function.addParameter('int_in1', dtype='int32', direction=function.IN, unit=units.m)
        function.addParameter('int_in2', dtype='int32', direction=function.IN, default=1, unit=units.kg)
        function.addParameter('int_out1', dtype='int32', direction=function.OUT, unit=units.m)
        function.addParameter('int_out2', dtype='int32', direction=function.OUT, unit=units.kg)
        function.addParameter('len', dtype='int32', direction=function.LENGTH)
        function.result_type = 'int32'
        function.must_handle_array = True
        return function

    @legacy_function
    def get_x():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('x', dtype='float32', direction=function.OUT, unit=units.m)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def set_x():
        function = LegacyFunctionSpecification()
        function.addParameter('index', dtype='int32', direction=function.IN)
        function.addParameter('x', dtype='float32', direction=function.IN, unit=units.m)
        function.result_type = 'int32'
        function.can_handle_array = True
        return function

    @legacy_function
    def dummy():
        function = LegacyFunctionSpecification()
        function.result_type = 'int32'
        return function


class ForTesting(InCodeComponentImplementation):

    def __init__(self, exefile, **options):
        InCodeComponentImplementation.__init__(self, ForTestingInterface(exefile, **options), **options)

    def get_grid_range(self):
        return (0, 9)

    def define_grids(self, handler):
        handler.define_grid('grid')
        handler.set_grid_range('grid', 'get_grid_range')
        handler.add_getter('grid', 'get_x', names=["x"])
        handler.add_setter('grid', 'set_x', names=["x"])


class ForTestingWithState(ForTesting):
    def define_state(self, handler):
        handler.set_initial_state("1")
        handler.add_transition("1", "2", "dummy")
        handler.add_method("2", "get_x")
        handler.add_method("2", "set_x")


class TestASync(TestWithMPI):

    @classmethod
    def setup_class(cls):
        print("building...")
        cls.check_can_compile_modules()
        try:
            cls.exefile = compile_tools.build_worker(codestring, cls.get_path_to_results(), ForTestingInterface)
        except Exception as ex:
            print(ex)
            raise
        print("done")

    def test1(self):
        instance = ForTestingInterface(self.exefile)
        int_out, error = instance.echo_int(10)
        instance.stop()
        self.assertEqual(int_out, 10)
        self.assertEqual(error, 0)

    def test2(self):
        instance = ForTestingInterface(self.exefile)
        request = instance.echo_int.asynchronous(10)
        self.assertEqual(request, instance.async_request)
        request.wait()
        int_out, error = request.result()
        self.assertEqual(int_out, 10)
        self.assertEqual(error, 0)
        instance.stop()

    def test3(self):
        instance = ForTestingInterface(self.exefile)
        request1 = instance.do_sleep.asynchronous(1)
        request2 = instance.echo_int.asynchronous(10)
        self.assertEqual(request2, instance.async_request)
        request2.wait()
        int_out, error = request2.result()
        self.assertEqual(int_out, 10)
        self.assertEqual(error, 0)
        instance.stop()

    def test4(self):
        instance = ForTesting(self.exefile)
        request1 = instance.do_sleep(1, return_request=True)
        request2 = instance.echo_int(10, return_request=True)
        self.assertEqual(request2, instance.async_request)
        instance.async_request.wait()
        int_out = request2.result()
        self.assertEqual(int_out, 10)
        instance.stop()

    def test5(self):
        instance = ForTesting(self.exefile)
        instance.do_sleep(1, return_request=True)
        requests = []
        for x in range(10):
            requests.append(instance.echo_int(x, return_request=True))
        instance.async_request.wait()
        for i, x in enumerate(requests):
            self.assertEqual(x.result(), i)
        instance.stop()

    def test6(self):
        instance = ForTesting(self.exefile)
        requests = []
        for x in range(10):
            requests.append(instance.echo_int(x, return_request=True))
        instance.async_request.wait()
        for i, x in enumerate(requests):
            self.assertEqual(x.result(), i)
        instance.stop()

    def test7(self):
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)
        t1 = time.time()

        requests = []
        for x in range(10):
            requests.append([instance1.echo_int(x, return_request=True), x])
        for x in range(10):
            requests.append([instance2.echo_int(x, return_request=True), x])

        instance1.do_sleep(1, return_request=True)
        instance2.do_sleep(1, return_request=True)

        pool = instance1.async_request.join(instance2.async_request)
        pool.waitall()
        t2 = time.time()

        for x in requests:
            self.assertEqual(x[0].result(), x[1])
        instance1.stop()
        instance2.stop()
        self.assertTrue(t2-t1 < 2.)

    def test8(self):
        from threading import Thread
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)
        t1 = time.time()

        requests = []
        for x in range(10):
            requests.append([instance1.echo_int(x, return_request=True), x])
        for x in range(10):
            requests.append([instance2.echo_int(x, return_request=True), x])

        instance1.do_sleep(1, return_request=True)
        instance2.do_sleep(1, return_request=True)

        pool = instance1.async_request.join(instance2.async_request)

        thread = Thread(target=pool.waitall)
        thread.start()
        time.sleep(1)
        thread.join()

        self.assertTrue(pool)

        t2 = time.time()

        for x in requests:
            self.assertEqual(x[0].result(), x[1])
        instance1.stop()
        instance2.stop()
        self.assertTrue(t2-t1 < 2.)

    def test9(self):
        instance = ForTesting(self.exefile)
        for x in range(10):
            instance.echo_int(x, return_request=True)
        results = instance.async_request.results
        self.assertEqual(results, list(range(10)))
        instance.stop()

    def test10(self):
        instance = ForTesting(self.exefile)
        r1 = instance.do_sleep(1, return_request=True)
        r2 = instance.return_error(return_request=True)
        r3 = instance.echo_int(1, return_request=True)
        r4 = instance.echo_int(2, return_request=True)
        self.assertRaises(Exception, instance.async_request.waitall,
            expected_message="Error in dependent call: Error when calling 'return_error' of a 'ForTesting', errorcode is -1")
        self.assertTrue(r1.is_result_available())
        self.assertFalse(r2.is_result_available())
        self.assertTrue(r2.is_finished)
        self.assertTrue(r3.is_finished)
        self.assertFalse(bool(r3))
        self.assertTrue(r4.is_finished)
        self.assertTrue(r4.waits_for() is None)
        self.assertFalse(r3.is_result_available())

        instance.stop()

    def test11(self):
        """ cross dependency """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_int(10, return_request=True)

        def fac():
            return instance2.echo_int(20, return_request=True)

        # ~ request2=instance2.echo_int(20, async_dependency=request1)
        request2 = async_request.DependentASyncRequest(request1, fac)

        request2.wait()

        self.assertEqual(request2.result(), 20)

        instance1.stop()
        instance2.stop()

    def test11b(self):
        """ cross dependency """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_int(10, return_request=True)
        request2 = instance2.echo_int(20, async_dependency=request1, return_request=True)

        request2.wait()
        self.assertTrue(request1.is_result_available())

        self.assertEqual(request2.result(), 20)

        instance1.stop()
        instance2.stop()

    def test12(self):
        """ cross dependency with input-output dependency """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_int(10, return_request=True)

        results = dict()

        def safe_result(arg, index):
            result = arg()
            results[index] = result
            return result

        request1.add_result_handler(safe_result, (1,))

        def fac():
            return instance2.echo_int(results[1], return_request=True)

        # ~ request2=instance2.echo_int(??, async_factory=fac)
        request2 = async_request.DependentASyncRequest(request1, fac)

        request2.wait()

        self.assertEqual(request2.result(), 10)

        instance1.stop()
        instance2.stop()

    def test12b(self):
        """ cross dependency with input-output dependency """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_int(10, return_request=True)

        request2 = instance2.echo_int(request1, return_request=True)

        request2.wait()

        self.assertEqual(request2.result(), 10)

        instance1.stop()
        instance2.stop()

    def test12c(self):
        """ cross dependency with input-output dependency """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_2_int(1 | units.m, 2 | units.kg, return_request=True)

        request2 = instance2.echo_2_int(request1[0], request1[1], return_request=True)

        print("do...wait...")
        request2.wait()
        print("done", request2.result())

        self.assertEqual(request2.result()[0], 1 | units.m)
        self.assertEqual(request2.result()[1], 2 | units.kg)

        instance1.stop()
        instance2.stop()

    def test12c(self):
        """ cross dependency with input-output dependency """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)
        instance3 = ForTesting(self.exefile)
        instance4 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_2_int(1 | units.m, 2 | units.kg, return_request=True)
        request1b = instance1.do_sleep(1, return_request=True)
        request2 = instance2.echo_2_int(3 | units.m, 4 | units.kg, return_request=True)
        request3 = instance3.echo_2_int(request2[0], 5 | units.kg, return_request=True)

        instance4.do_sleep(1, return_request=True)
        request4 = instance4.echo_2_int(request2[0], request3[1], return_request=True, async_dependency=request1b)

        request3.wait()

        self.assertEqual(request4.result()[0], 3 | units.m)
        self.assertEqual(request4.result()[1], 5 | units.kg)

        instance1.stop()
        instance2.stop()
        instance3.stop()
        instance4.stop()

    def test13(self):
        instance = ForTesting(self.exefile)

        r = instance.echo_int(1, return_request=True)
        time.sleep(0.1)
        self.assertTrue(r.is_result_available())
        r.result()

        r = instance.return_error(return_request=True)
        time.sleep(0.1)
        self.assertTrue(r.is_result_available())
        self.assertTrue(r.is_result_available())
        self.assertRaises(Exception, r.result, expected_message="Error when calling 'return_error' of a 'ForTesting', errorcode is -1")
        self.assertFalse(r.is_result_available())
        self.assertTrue(r.is_finished)
        instance.stop()

    def test14(self):
        instance = ForTesting(self.exefile, channel_type="sockets")

        r = instance.echo_int(1, return_request=True)
        time.sleep(0.1)
        self.assertTrue(r.is_result_available())
        r.result()

        r = instance.return_error(return_request=True)
        time.sleep(0.1)
        self.assertTrue(r.is_result_available())
        self.assertTrue(r.is_result_available())
        self.assertRaises(Exception, r.result, expected_message="Error when calling 'return_error' of a 'ForTesting', errorcode is -1")
        self.assertFalse(r.is_result_available())
        self.assertTrue(r.is_finished)
        instance.stop()

    def test15(self):
        instance = ForTesting(self.exefile)

        instance.do_sleep(1, return_request=True)
        instance.return_error(return_request=True)
        instance.echo_int(1, return_request=True)
        instance.echo_int(1, return_request=True)
        # ~ self.assertRinstance.echo_int(1)

        instance.stop()

    def test16(self):
        instance = ForTesting(self.exefile)

        instance.do_sleep(1, return_request=True)
        result = instance.echo_2_int([11, 12, 13] | units.m, [3, 2, 1] | units.kg, return_request=True)

        r1 = result[0]
        r2 = result[1]

        self.assertEqual(r1.result(), [11, 12, 13] | units.m)
        self.assertEqual(r2.result(), [3, 2, 1] | units.kg)

        instance.stop()

    def test17(self):
        instance = ForTestingInterface(self.exefile)

        instance.do_sleep.asynchronous(1)
        request = instance.echo_2_int.asynchronous([11, 12, 13], [3, 2, 1])

        r1 = request["int_out1"]
        r2 = request["int_out2"]

        self.assertEqual(r1.result(), [11, 12, 13])
        self.assertEqual(r2.result(), [3, 2, 1])

        instance.stop()

    def test18(self):
        """ test pool as depedency 1 """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)
        instance3 = ForTesting(self.exefile)

        request0 = instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_int(10, return_request=True)
        request2 = instance2.echo_int(10, return_request=True)
        request = async_request.AsyncRequestsPool(request1, request2)

        request3 = instance3.echo_int(11, async_dependency=request, return_request=True)
        request3.wait()
        self.assertTrue(request1.is_result_available())
        self.assertTrue(request2.is_result_available())
        self.assertEqual(request3.result(), 11)

        instance1.stop()
        instance2.stop()
        instance3.stop()

    def test18b(self):
        """ test pool as depedency 2 """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)
        instance3 = ForTesting(self.exefile)

        request0 = instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_int(10, return_request=True)
        request2 = instance1.echo_int(10, return_request=True)
        request = async_request.AsyncRequestsPool(request1, request2)

        request3 = instance3.echo_int(11, async_dependency=request, return_request=True)
        request3.wait()
        self.assertTrue(request1.is_result_available())
        self.assertTrue(request2.is_result_available())
        self.assertEqual(request3.result(), 11)

        instance1.stop()
        instance2.stop()
        instance3.stop()

    def test19(self):
        """ test sum request """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)

        r1 = instance1.echo_int(1, return_request=True)
        r2 = instance2.echo_int(2, return_request=True)
        s = r1+r2
        r1 = instance1.echo_int(2, return_request=True)
        r2 = instance2.echo_int(3, return_request=True)
        m = r1*r2
        r1 = instance1.echo_int(12, return_request=True)
        r2 = instance2.echo_int(3, return_request=True)
        d = r1/r2

        self.assertEqual(s.result(), 3)
        self.assertEqual(m.result(), 6)
        self.assertEqual(d.result(), 4)

        instance1.stop()
        instance2.stop()

    def test19b(self):
        """ test sum request """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)

        r1 = instance1.echo_int(1, return_request=True)
        s = r1+2
        r1 = instance1.echo_int(2, return_request=True)
        m = r1*3
        r2 = instance2.echo_int(3, return_request=True)
        d = 12/r2

        self.assertEqual(s.result(), 3)
        self.assertEqual(m.result(), 6)
        self.assertEqual(d.result(), 4)

        instance1.stop()
        instance2.stop()

    def test20(self):
        """ some more tests of request expressions """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)
        instance3 = ForTesting(self.exefile)

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_2_int(1 | units.m, 2 | units.kg, return_request=True)
        request2 = instance2.echo_2_int(4 | units.m, 6 | units.kg, return_request=True)
        request3a = instance3.echo_int(request2[0] / request1[0]-4, return_request=True)
        request3b = instance3.echo_int(request2[1] / request1[1]-3, return_request=True)

        request3a.wait()
        request3b.wait()

        self.assertEqual(request3a.result(), 0)
        self.assertEqual(request3b.result(), 0)

        instance1.stop()
        instance2.stop()
        instance3.stop()

    def test21(self):
        """ test sum request, sockets """
        instance1 = ForTesting(self.exefile, channel_type="sockets")
        instance2 = ForTesting(self.exefile, channel_type="sockets")

        instance1.do_sleep(1, return_request=True)

        r1 = instance1.echo_int(1, return_request=True)
        r2 = instance2.echo_int(2, return_request=True)
        s = r1+r2
        r1 = instance1.echo_int(2, return_request=True)
        r2 = instance2.echo_int(3, return_request=True)
        m = r1*r2
        r1 = instance1.echo_int(12, return_request=True)
        r2 = instance2.echo_int(3, return_request=True)
        d = r1/r2

        self.assertEqual(s.result(), 3)
        self.assertEqual(m.result(), 6)
        self.assertEqual(d.result(), 4)

        instance1.stop()
        instance2.stop()

    def test21(self):
        """ some more tests of request expressions """
        instance1 = ForTesting(self.exefile)

        a = [10, 30, 15] | units.m
        b = [1, 3, 5] | units.kg

        instance1.do_sleep(1, return_request=True)
        request1 = instance1.echo_2_int(a, b, return_request=True)

        request2 = (3*request1[1]/(2.*request1[0])+(55. | units.kg/units.m))
        self.assertEqual(request2.result(), (3*b/(2.*a)+(55. | units.kg/units.m)))

        instance1.stop()

    def test22(self):
        """ tests of unpack """
        instance1 = ForTesting(self.exefile)

        a = [10, 30, 15] | units.m
        b = [1, 3, 5] | units.kg

        # ~ instance1.do_sleep(1, return_request=True)
        a_, b_ = instance1.echo_2_int(a, b, return_request=True)

        self.assertEqual((3*b_/(2.*a_)+(55. | units.kg/units.m)).result(), (3*b/(2.*a)+(55. | units.kg/units.m)))

        instance1.stop()

    def test23(self):
        """ tests of unpack """
        instance1 = ForTestingInterface(self.exefile)

        a = [10, 30, 15]
        b = [1, 3, 5]

        # ~ instance1.do_sleep(1, return_request=True)
        res = instance1.echo_2_int.asynchronous(a, b)
        # ~ res=res['int_out1']
        a_, b_, err = res

        self.assertEqual(a, a_.result())
        self.assertEqual(b, b_.result())

        instance1.stop()

    def test24(self):
        """ more test of pool """
        instance1 = ForTesting(self.exefile)
        instance2 = ForTesting(self.exefile)

        r1 = instance1.echo_int(1, return_request=True)
        r2 = instance1.echo_int(2, return_request=True)
        r3 = instance2.echo_int(3, return_request=True)
        r4 = instance2.echo_int(4, return_request=True)

        p1 = r1.join(r3)
        p2 = r2.join(r4)

        p3 = p1.join(p2)

        self.assertTrue(p3 is p1)

        p3.waitall()

        self.assertEqual(r1.result(), 1)
        self.assertEqual(r2.result(), 2)
        self.assertEqual(r3.result(), 3)
        self.assertEqual(r4.result(), 4)

        instance1.stop()
        instance2.stop()

    def test25(self):
        """ more test of pool: calls of same code """
        from amuse.rfi.async_request import AsyncRequestsPool
        instance1 = ForTesting(self.exefile)

        r1 = instance1.do_sleep(1, return_request=True)
        r2 = instance1.echo_int(2, return_request=True)

        p1 = AsyncRequestsPool()
        r1.wait()
        r2.wait()
        p1.add_request(r1)
        p1.add_request(r2)

        # ~ p1=r1.join(r2)

        p1.waitall()

        self.assertEqual(r2.result(), 2)

        instance1.stop()

    def test26(self):
        """ test a lot of calls"""
        instances = []
        for i in range(20):
            instances.append(ForTesting(self.exefile, channel_type="sockets"))

        import time

        print("ready")
        time.sleep(1)

        print("start")
        t1 = time.time()
        pool = []
        for i in instances:
            print(i, "submit")
            pool.append(i.do_sleep(5, return_request=True))

        for i, p in enumerate(pool):
            print("result:", i, p.result())
        t2 = time.time()
        print(t2-t1)
        for i in instances:
            i.stop()

    def test30(self):
        """ test a grid attribute request """
        instance1 = ForTesting(self.exefile, redirection="none")
        self.assertEqual(instance1.grid.x, numpy.arange(1, 11) | units.m)
        instance1.do_sleep(1, return_request=True)
        t1 = time.time()
        request = instance1.grid.request.x
        t2 = time.time()
        self.assertLess(t2-t1, 0.5)
        self.assertEqual(request.result(), numpy.arange(1, 11) | units.m)
        t2 = time.time()
        self.assertGreater(t2-t1, 1.)

    def test31(self):
        """ test a grid attribute request, subgrids """
        instance1 = ForTesting(self.exefile, redirection="none")
        self.assertEqual(instance1.grid.x, numpy.arange(1, 11) | units.m)
        instance1.do_sleep(1, return_request=True)
        t1 = time.time()
        request = instance1.grid[:5].request.x
        request2 = instance1.grid[5:].request.x
        t2 = time.time()
        self.assertLess(t2-t1, 0.5)
        self.assertEqual(request.result(), numpy.arange(1, 6) | units.m)
        self.assertEqual(request2.result(), numpy.arange(6, 11) | units.m)
        t2 = time.time()
        self.assertGreater(t2-t1, 1.)

    def test32(self):
        """ test a grid attribute request setter """
        instance1 = ForTesting(self.exefile, redirection="none")
        instance1.grid.x = (66.+numpy.arange(1, 11)) | units.m
        self.assertEqual(instance1.grid.x, (66.+numpy.arange(1, 11)) | units.m)

        t1 = time.time()
        instance1.do_sleep(1, return_request=True)
        instance1.grid.request.x = (11.+numpy.arange(1, 11)) | units.m
        t2 = time.time()
        self.assertLess(t2-t1, 0.5)
        instance1.async_request.wait()
        t2 = time.time()
        self.assertGreater(t2-t1, 1.)
        t1 = time.time()
        self.assertEqual(instance1.grid.x, (11.+numpy.arange(1, 11)) | units.m)
        t2 = time.time()
        self.assertLess(t2-t1, 0.5)

    def test33(self):
        """ test a grid attribute request, subgrids """
        instance1 = ForTesting(self.exefile, redirection="none")
        self.assertEqual(instance1.grid.x, numpy.arange(1, 11) | units.m)

        t1 = time.time()
        instance1.do_sleep(1, return_request=True)
        instance1.grid[::2].request.x = (11.+numpy.arange(1, 11, 2)) | units.m
        t2 = time.time()
        self.assertLess(t2-t1, 0.5)
        instance1.async_request.wait()
        t2 = time.time()
        self.assertGreater(t2-t1, 1.)
        self.assertEqual(instance1.grid.x[::2], (11.+numpy.arange(1, 11, 2)) | units.m)
        self.assertEqual(instance1.grid.x[1::2], (numpy.arange(2, 11, 2)) | units.m)

    def test34(self):
        """ test a grid attribute request, subgrids """
        instance1 = ForTesting(self.exefile, redirection="none")
        grid = instance1.grid.copy()
        request = instance1.grid.request.x
        self.assertEqual(request.result(), numpy.arange(1, 11) | units.m)

    def test35(self):
        """ test a grid attribute request setter with state"""
        instance1 = ForTestingWithState(self.exefile, redirection="none")
        t1 = time.time()
        instance1.do_sleep(1, return_request=True)
        self.assertEqual(instance1.get_name_of_current_state(), '1')
        instance1.grid.request.x = (11.+numpy.arange(1, 11)) | units.m
        self.assertEqual(instance1.get_name_of_current_state(), '2')
        t2 = time.time()
        self.assertGreater(t2-t1, 1.)   # first time, state calls dummy (blocking) -> wait

        t1 = time.time()
        instance1.do_sleep(1, return_request=True)
        instance1.grid.request.x = (12.+numpy.arange(1, 11)) | units.m
        t2 = time.time()
        self.assertLess(t2-t1, 0.5)  # second time should be less

        instance1.async_request.wait()
        t2 = time.time()
        self.assertGreater(t2-t1, 1.)
        t1 = time.time()
        self.assertEqual(instance1.grid.x, (12. + numpy.arange(1, 11)) | units.m)
        t2 = time.time()
        self.assertLess(t2-t1, 0.5)

    def test36(self):
        """ more state tests"""
        instance1 = ForTestingWithState(self.exefile, redirection="none")
        self.assertEqual(instance1.get_name_of_current_state(), '1')
        # this documents current behaviour:
        instance1.dummy(return_request=True)
        self.assertEqual(instance1.get_name_of_current_state(), '1')
        instance1.async_request.wait()
        self.assertEqual(instance1.get_name_of_current_state(), '2')
        # ie state changes upon completion of call at wait. This is
        # sort of ok, alternatively state could be changed immediately...


class TestASyncDistributed(TestASync):

    @classmethod
    def setup_class(cls):
        cls.check_not_in_mpiexec()
        super(TestASyncDistributed, cls).setup_class()
        cls.distinstance = cls.new_instance_of_an_optional_code(DistributedAmuse)  # , redirection='none')
        cls.distinstance.parameters.debug = False

        # ~ print "Resources:"
        # ~ print cls.distinstance.resources

        pilot = Pilot()
        pilot.resource_name = 'local'
        pilot.node_count = 1
        pilot.time = 2 | units.hour
        pilot.slots_per_node = 8
        pilot.label = 'local'
        cls.distinstance.pilots.add_pilot(pilot)
        # ~ print "Pilots:"
        # ~ print cls.distinstance.pilots

        # ~ print "Waiting for pilots"
        cls.distinstance.wait_for_pilots()
        cls.distinstance.use_for_all_workers()

    @classmethod
    def tearDown(cls):
        # ~ print "Stopping distributed code"
        cls.distinstance.stop()

    @classmethod
    def check_not_in_mpiexec(cls):
        """
        The tests will fork another process, if the test run
        is itself an mpi process, the tests may fail. 

        For the hydra process manager the tests will fail.
        So skip the tests if we detect hydra
        """
        if 'HYDI_CONTROL_FD' in os.environ:
            return
        if 'HYDRA_CONTROL_FD' in os.environ or 'PMI_FD' in os.environ:
            cls.skip('cannot run the socket tests under hydra process manager')


if __name__ == "__main__":
    TestASync.setup_class()
    t = TestASync()

    t.test26()
